# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
import random
import math
import warnings
# The classes below are unused, but imported so that they are available in the
# workspace.
from openexp.canvas_elements import (Rect, Line, Text, Ellipse, Circle,
                                     FixDot, Gabor, NoisePatch, Image, Arrow,
                                     Polygon)
from libopensesame.widgets.widget_factory import (Label, Button, ImageWidget,
                                                  ImageButton, TextInput,
                                                  RatingScale, Checkbox)


# Factory functions
def Experiment(osexp_path=None, log_path='defaultlog.csv', fullscreen=False,
               subject_nr=0, **kwargs):
    """A factory function that creates a new `Experiment` object. This is only
    useful when implementing an experiment entirely through a Python script,
    rather than through the user interface.
    
    Parameters
    ----------
    osexp_path : str or Path or None, optional
        If a path to an `.osexp` file is specified, this file is opened and
        can be run directly by calling `Experiment.run()`. If no experiment
        file is specified, an empty experiment is initialized.
    log_path : str or Path, optional
    fullscreen : bool, optional
    subject_nr : int, optional
    kwargs : dict, optional
        Optional keyword arguments that are interpreted as experimental
        variables. The main use of this is to specify the backend through
        the `canvas_backend` keyword.
        
    Examples
    --------
    To implement an experiment fully programmatically:
    
    ~~~ .python
    from libopensesame.python_workspace_api import (
        Experiment, Canvas, Text, Keyboard)
    exp, win, clock, log = Experiment(canvas_backend='legacy')
    c = Canvas()
    c += Text('Press any key')
    c.show()
    kb = Keyboard()
    kb.get_key()
    exp.end()
    ~~~
    
    To load an experiment file and run it:

    ~~~ .python
    from libopensesame.python_workspace_api import Experiment
    exp, win, clock, log = Experiment(osexp_path='my_experiment.osexp',
                                      subject_nr=2)
    exp.run()
    ~~~
    
    Returns
    -------
    tuple
        An (exp, win, clock, log) tuple corresponding to the Experiment,
        window handle (backend-specific), Clock, and Log objects.
    """
    global experiment
    from libopensesame.experiment import Experiment as RuntimeExperiment
    if osexp_path is None:
        from libopensesame.syntax import Syntax
        syntax = Syntax(None)
        string = '\n'.join([syntax.create_cmd('set', [key, value])
                            for key, value in kwargs.items()])
    else:
        string = osexp_path
    experiment = RuntimeExperiment(experiment_path=osexp_path,
                                   logfile=log_path, fullscreen=fullscreen,
                                   subject_nr=subject_nr, string=string)
    experiment.init_random()
    experiment.init_display()
    experiment.init_clock()
    experiment.init_sound()
    experiment.init_log()
    return experiment, experiment.window, experiment.clock, experiment.log


def Form(*args, **kwargs):
    r"""A factory function that creates a new `Form` object. For a
    description
    of possible keywords, see:

    - %link:manual/forms/widgets%

    Returns
    -------
    canvas
        A `Form` object.

    Examples
    --------
    >>> form = Form()
    >>> label = Label(text='label')
    >>> button = Button(text='Ok')
    >>> form.set_widget(label, (0,0))
    >>> form.set_widget(button, (0,1))
    >>> form._exec()
    """
    from libopensesame.widgets import form
    return form(experiment, **kwargs)


def Canvas(auto_prepare=True, **style_args):
    r"""A factory function that creates a new `Canvas` object. For a
    description of possible keywords, see:

    - %link:manual/python/canvas%

    Returns
    -------
    canvas
        A `Canvas` object.

    Examples
    --------
    >>> my_canvas = Canvas(color=u'red', penwidth=2)
    >>> my_canvas.line(-10, -10, 10, 10)
    >>> my_canvas.line(-10, 10, 10, -10)
    >>> my_canvas.show()
    """
    from openexp.canvas import Canvas
    return Canvas(experiment, auto_prepare=auto_prepare, **style_args)


def Keyboard(**resp_args):
    r"""A factory function that creates a new `Keyboard` object. For a
    description of possible keywords, see:

    - %link:manual/python/keyboard%

    Returns
    -------
    keyboard
        A `Keyboard` object.

    Examples
    --------
    >>> my_keyboard = Keyboard(keylist=[u'a', u'b'], timeout=5000)
    >>> key, time = my_keyboard.get_key()
    """
    from openexp.keyboard import Keyboard
    return Keyboard(experiment, **resp_args)


def Mouse(**resp_args):
    r"""A factory function that creates a new `Mouse` object. For a
    description
    of possible keywords, see:

    - %link:manual/python/mouse%

    Returns
    -------
    mouse
        A `mouse` object.

    Examples
    --------
    >>> my_mouse = Mouse(keylist=[1,3], timeout=5000)
    >>> button, time = my_mouse.get_button()
    """
    from openexp.mouse import Mouse
    return Mouse(experiment, **resp_args)


def Sampler(src, **playback_args):
    r"""A factory function that creates a new `Sampler` object. For a
    description of possible keywords, see:

    - %link:manual/python/sampler%

    Returns
    -------
    sampler
        A SAMPLER object.

    Examples
    --------
    >>> src = pool['bark.ogg']
    >>> my_sampler = Sampler(src, volume=.5, pan='left')
    >>> my_sampler.play()
    """
    from openexp.sampler import Sampler
    return Sampler(experiment, src, **playback_args)


# Miscellaneous API	functions


def Synth(osc="sine", freq=440, length=100, attack=0, decay=5):
    r"""A factory function that synthesizes a sound and returns it as a
    `Sampler` object.

    Parameters
    ----------
    osc : str, unicode, optional
        Oscillator, can be "sine", "saw", "square" or "white_noise".
    freq : str, unicode, int, float, optional
        Frequency, either an integer value (value in hertz) or a string ("A1",
        "eb2", etc.).
    length : int, float, optional
        The length of the sound in milliseconds.
    attack : int, float, optional
        The attack (fade-in) time in milliseconds.
    decay : int, float, optional
        The decay (fade-out) time in milliseconds.

    Returns
    -------
    sampler
        A SAMPLER object.

    Examples
    --------
    >>> my_sampler = Synth(freq=u'b2', length=500)
    """
    from openexp.synth import Synth
    return Synth(experiment, osc=osc, freq=freq, length=length, attack=attack,
                 decay=decay)


def copy_sketchpad(name):
    r"""Returns a copy of a `sketchpad`'s canvas.

    Parameters
    ----------
    name : str, unicode
        The name of the `sketchpad`.

    Returns
    -------
    canvas
        A copy of the `sketchpad`'s canvas.

    Examples
    --------
    >>> my_canvas = copy_sketchpad('my_sketchpad')
    >>> my_canvas.show()
    """
    c = Canvas()
    c.copy(experiment.items[name].canvas)
    return c


def reset_feedback():
    r"""Resets all feedback variables to their initial state.

    Examples
    --------
    >>> reset_feedback()
    """
    experiment.reset_feedback()


def set_subject_nr(nr):
    r"""Sets the subject number and parity (even/ odd). This function is called
    automatically when an experiment is started, so you only need to call it
    yourself if you overwrite the subject number that was specified when the
    experiment was launched.

    Parameters
    ----------
    nr : int
        The subject nr.

    Examples
    --------
    >>> set_subject_nr(1)
    >>> print('Subject nr = %d' % var.subject_nr)
    >>> print('Subject parity = %s' % var.subject_parity)
    """
    experiment.set_subject(nr)


def sometimes(p=.5):
    r"""Returns True with a certain probability. (For more advanced
    randomization, use the Python `random` module.)

    Parameters
    ----------
    p : float, optional
        The probability of returning True.

    Returns
    -------
    bool
        True or False

    Examples
    --------
    >>> if sometimes():
    >>>         print('Sometimes you win')
    >>> else:
    >>>         print('Sometimes you loose')
    """
    if (not isinstance(p, float) and not isinstance(p, int)) or p < 0 or p > 1:
        raise ValueError(
            f'p should be a numeric value between 0 and 1, not {p}')
    return random.random() < p


def pause():
    r"""Pauses the experiment."""
    experiment.pause()


def register_cleanup_function(fnc):
    r"""Registers a clean-up function, which is executed when the experiment
    ends. Clean-up functions are executed at the very end, after the display,
    sound device, and log file have been closed. Clean-up functions are also
    executed when the experiment crashes.

    Examples
    --------
    >>> def my_cleanup_function():
    >>>         print(u'The experiment is finished!')
    >>> register_cleanup_function(my_cleanup_function)
    """
    experiment.cleanup_functions.append(fnc)


def xy_from_polar(rho, phi, pole=(0, 0)):
    r"""Converts polar coordinates (distance, angle) to Cartesian coordinates
    (x, y).

    Parameters
    ----------
    rho : float
        The radial coordinate, also distance or eccentricity.
    phi : float
        The angular coordinate. This reflects a clockwise rotation in degrees
        (i.e. not radians), where 0 is straight right.
    pole : tuple, optional
        The reference point.

    Returns
    -------
    tuple
        An (x, y) coordinate tuple.

    Examples
    --------
    >>> # Draw a cross
    >>> x1, y1 = xy_from_polar(100, 45)
    >>> x2, y2 = xy_from_polar(100, -45)
    >>> c = Canvas()
    >>> c.line(x1, y1, -x1, -y1)
    >>> c.line(x2, y2, -x2, -y2)
    >>> c.show()
    """
    try:
        rho = float(rho)
    except:
        raise TypeError('rho should be numeric in xy_from_polar()')
    try:
        phi = float(phi)
    except:
        raise TypeError('phi should be numeric in xy_from_polar()')
    phi = math.radians(phi)
    ox, oy = parse_pole(pole)
    x = rho * math.cos(phi) + ox
    y = rho * math.sin(phi) + oy
    return x, y


def xy_to_polar(x, y, pole=(0, 0)):
    r"""Converts Cartesian coordinates (x, y) to polar coordinates (distance,
    angle).

    Parameters
    ----------
    x : float
        The X coordinate.
    y : float
        The Y coordinate.
    pole : tuple, optional
        The reference point.

    Returns
    -------
    tuple
        An (rho, phi) coordinate tuple. Here, `rho` is the radial coordinate,
        also distance or eccentricity. `phi` is the angular coordinate in
        degrees (i.e. not radians), and reflects a counterclockwise rotation,
        where 0 is straight right.

    Examples
    --------
    >>> rho, phi = xy_to_polar(100, 100)
    """
    try:
        x = float(x)
    except:
        raise TypeError('x should be numeric in xy_to_polar()')
    try:
        y = float(y)
    except:
        raise TypeError('y should be numeric in xy_to_polar()')
    ox, oy = parse_pole(pole)
    dx = x-ox
    dy = y-oy
    rho = math.sqrt(dx**2 + dy**2)
    phi = math.degrees(math.atan2(dy, dx))
    return rho, phi


def xy_distance(x1, y1, x2, y2):
    r"""Gives the distance between two points.

    Parameters
    ----------
    x1 : float
        The x coordinate of the first point.
    y1 : float
        The y coordinate of the first point.
    x2 : float
        The x coordinate of the second point.
    y2 : float
        The y coordinate of the second point.

    Returns
    -------
    float
        The distance between the two points.
    """
    try:
        x1 = float(x1)
        y1 = float(y1)
        x2 = float(x2)
        y2 = float(y2)
    except:
        raise TypeError('Coordinates should be numeric in xy_distance()')
    return math.sqrt((x1-x2)**2+(y1-y2)**2)


def xy_circle(n, rho, phi0=0, pole=(0, 0)):
    r"""Generates a list of points (x,y coordinates) in a circle. This can be
    used to draw stimuli in a circular arrangement.

    Parameters
    ----------
    n : int
        The number of x,y coordinates to generate.
    rho : float
        The radial coordinate, also distance or eccentricity, of the first
        point.
    phi0 : float, optional
        The angular coordinate for the first coordinate. This is a
        counterclockwise rotation in degrees (i.e. not radians), where 0 is
        straight right.
    pole : tuple, optional
        The reference point.

    Returns
    -------
    list
        A list of (x,y) coordinate tuples.

    Examples
    --------
    >>> # Draw 8 rectangles around a central fixation dot
    >>> c = Canvas()
    >>> c.fixdot()
    >>> for x, y in xy_circle(8, 100):
    >>>         c.rect(x-10, y-10, 20, 20)
    >>> c.show()
    """
    try:
        n = int(n)
        if n < 0:
            raise ValueError()
    except (ValueError, TypeError):
        raise ValueError('n should be a non-negative integer in xy_circle()')
    try:
        phi0 = float(phi0)
    except (ValueError, TypeError):
        raise TypeError('phi0 should be numeric in xy_circle()')
    l = []
    for i in range(n):
        l.append(xy_from_polar(rho, phi0, pole=pole))
        phi0 += 360./n
    return l


def xy_grid(n, spacing, pole=(0, 0)):
    r"""Generates a list of points (x,y coordinates) in a grid. This can be
    used to draw stimuli in a grid arrangement.

    Parameters
    ----------
    n : int, tuple
        An `int` that indicates the number of columns and rows, so that `n=2`
        indicates a 2x2 grid, or a (n_col, n_row) `tuple`, so that `n=(2,3)`
        indicates a 2x3 grid.
    spacing : float
        A numeric value that indicates the spacing between cells, or a
        (col_spacing, row_spacing) tuple.
    pole : tuple, optional
        The reference point.

    Returns
    -------
    list
        A list of (x,y) coordinate tuples.

    Examples
    --------
    >>> # Draw a 4x4 grid of rectangles
    >>> c = Canvas()
    >>> c.fixdot()
    >>> for x, y in xy_grid(4, 100):
    >>>         c.rect(x-10, y-10, 20, 20)
    >>> c.show()
    """
    try:
        n_col, n_row = n
    except:
        n_col = n_row = n
    try:
        n_col = int(n_col)
        n_row = int(n_row)
        assert(n_col >= 0)
        assert(n_row >= 0)
    except:
        raise ValueError('n should be a non-negative integer or a tuple of '
                         'two non-negative integers in xy_grid()')
    try:
        s_col, s_row = spacing
    except:
        s_col = s_row = spacing
    try:
        s_col = float(s_col)
        s_row = float(s_row)
        assert(s_col >= 0)
        assert(s_row >= 0)
    except:
        raise ValueError('spacing should be a non-negative numeric or a '
                         'tuple of two non-negative numerics in xy_grid()')
    pole = parse_pole(pole)
    l = []
    for row in range(n_row):
        y = (row - (n_row-1) / 2.) * s_row + pole[1]
        for col in range(n_col):
            x = (col - (n_col-1) / 2.) * s_col + pole[0]
            l.append((x, y))
    return l


def xy_random(n, width, height, min_dist=0, pole=(0, 0)):
    r"""Generates a list of random points (x,y coordinates) with a minimum
    spacing between each pair of points. This function will raise an
    Exception when the coordinate list cannot be generated,  typically because
    there are too many points, the min_dist is set too high, or the width or
    height are set too low.

    Parameters
    ----------
    n : int
        The number of points to generate.
    width : float
        The width of the field with random points.
    height : float
        The height of the field with random points.
    min_dist : float, optional
        The minimum distance between each point.
    pole : tuple, optional
        The reference point.

    Returns
    -------
    list
        A list of (x,y) coordinate tuples.

    Examples
    --------
    >>> # Draw a 50 rectangles in a random grid
    >>> c = Canvas()
    >>> c.fixdot()
    >>> for x, y in xy_random(50, 500, 500, min_dist=40):
    >>>         c.rect(x-10, y-10, 20, 20)
    >>> c.show()
    """
    try:
        n = int(n)
        if n < 0:
            raise ValueError()
    except (ValueError, TypeError):
        raise ValueError('n should be a non-negative integer in xy_circle()')
    try:
        width = float(width)
    except:
        raise TypeError('width should be numeric in xy_random()')
    try:
        height = float(height)
    except:
        raise TypeError('height should be numeric in xy_random()')
    try:
        min_dist = float(min_dist)
    except:
        raise TypeError('min_dist should be numeric in xy_random()')
    pole = parse_pole(pole)
    max_try = 1000
    for t1 in range(max_try):
        l = []
        for i in range(n):
            for t2 in range(max_try):
                x1 = (random.random()-.5)*width + pole[0]
                y1 = (random.random()-.5)*height + pole[1]
                for x2, y2 in l:
                    if xy_distance(x1, y1, x2, y2) < min_dist:
                        break
                else:
                    # Point does not collide, so add to the list
                    l.append((x1, y1))
                    break
            else:
                # All level-2 tries have failed, so break to start a new level-1
                # try.
                break
        else:
            # All points have been successfully added, so return the list
            return l
    # All level-1 tries have failed
    raise RuntimeError('Failed to generate random coordinates in xy_random()')


# Helper functions that are not part of the public API


def parse_pole(pole):
    try:
        ox = float(pole[0])
        oy = float(pole[1])
        assert(len(pole) == 2)
    except:
        raise ValueError('pole should be a tuple (or similar) of length '
                         'with two numeric values')
    return ox, oy


def set_aliases():
    # Non PEP-8 alias for backwards compatibility
    global canvas, sampler, synth, keyboard, mouse
    canvas = Canvas
    sampler = Sampler
    synth = Synth
    keyboard = Keyboard
    mouse = Mouse



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from qtpy import QtCore, QtWidgets
from libqtopensesame.misc.base_subcomponent import BaseSubcomponent
from libqtopensesame.misc.translate import translation_context
_ = translation_context('item_category', category='core')


class ToolbarItemsLabel(BaseSubcomponent, QtWidgets.QFrame):
    """A label for the item toolbar.
    
    Parameters
    ----------
    parent : QWidget
        The parent.
    label : unicode
        Label text.
    """
    def __init__(self, parent, label):
        super().__init__(parent)
        self.setup(parent)
        l = QtWidgets.QLabel(_(label))
        l.setMaximumWidth(90)
        l.setIndent(6)
        l.setWordWrap(True)
        hbox = QtWidgets.QHBoxLayout()
        hbox.setContentsMargins(0, 0, 0, 0)
        hbox.addWidget(l)
        self.setLayout(hbox)


# Alias for backwards compatibility
toolbar_items_label = ToolbarItemsLabel



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame import misc, metadata
from libqtopensesame.misc.config import cfg
from libqtopensesame.widgets.base_widget import BaseWidget
from libopensesame.py3compat import *
from qtpy import QtWidgets, QtCore


class CreditsWidget(BaseWidget):

    r"""A credits window, embedded in the start-new widget."""
    def __init__(self, main_window):
        r"""Constructor.

        Parameters
        ----------
        main_window, optional
            A qtopensesame object.
        """
        super().__init__(main_window, ui=u'widgets.credits_widget')
        self.ui.label_opensesame.setText(self.ui.label_opensesame.text() % {
            u'version': metadata.__version__,
            u'codename': metadata.codename,
            u'channel': metadata.channel,
            u'python_version': metadata.python_version
        })
        self.ui.label_website.mousePressEvent = self.open_website
        self.ui.label_facebook.mousePressEvent = self.open_facebook
        self.ui.label_twitter.mousePressEvent = self.open_twitter

    def open_website(self, dummy=None):
        """Open the main website"""
        misc.open_url(cfg.url_website)

    def open_facebook(self, dummy=None):
        """Open Facebook page"""
        misc.open_url(cfg.url_facebook)

    def open_twitter(self, dummy=None):
        """Open Twitter page"""
        misc.open_url(cfg.url_twitter)


# Alias for backwards compatibility
credits_widget = CreditsWidget



"""A plug-in for using the serial response box."""

category = "Response collection"
help_url = 'manual/response/srbox'
controls = [
    {
        "label": "Dummy mode (use keyboard instead)",
        "var": "_dummy",
        "type": "checkbox",
        "tooltip": "Enable to respond with the keyboard instead of an SR Box"
    },
    {
        "label": "Ignore buttons that are already pressed",
        "var": "require_state_change",
        "type": "checkbox",
        "tooltip": "Require a button-state change, so that already-pressed buttons are ignored"
    },
    {
        "label": "Device name",
        "var": "dev",
        "type": "line_edit",
        "tooltip": "Expecting a valid device name. Leave empty for autodetect."
    },
    {
        "label": "Correct response",
        "var": "correct_response",
        "type": "line_edit",
        "tooltip": "Expecting a button number (1 .. 5)"
    },
    {
        "label": "Allowed responses",
        "var": "allowed_responses",
        "type": "line_edit",
        "tooltip": "Expecting a semicolon-separated list of button numbers, e.g., 1;3;4"
    },
    {
        "label": "Timeout",
        "var": "timeout",
        "type": "line_edit",
        "tooltip": "Expecting a value in milliseconds or 'infinite'",
        "validator": "timeout"
    },
    {
        "label": "Turn on lights",
        "var": "lights",
        "type": "line_edit",
        "tooltip": "Expecting a semicolon-separated list of light numbers, e.g., 1;3;4"
    },
    {
        "label": "<small><b>Note:</b> If there are multiple srbox items in the experiment, the first srbox item determines the device name and whether dummy mode is enabled</small>\n",
        "type": "text"
    }
]



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from libqtopensesame.extensions import BaseExtension
from libqtopensesame.misc.config import cfg
from .variable_inspector_dockwidget import VariableInspectorDockWidget
from qtpy import QtCore, QtGui, QtWidgets
from libqtopensesame.misc.translate import translation_context
_ = translation_context(u'variable_inspector', category=u'extension')


class VariableInspector(BaseExtension):

    def event_startup(self):
        self.need_refresh = False
        self.dock_widget = VariableInspectorDockWidget(
            self.main_window, self)
        self.dock_widget.visibilityChanged.connect(self.set_visible)
        self.main_window.addDockWidget(
            QtCore.Qt.RightDockWidgetArea,
            self.dock_widget
        )
        self.set_visible(cfg.variable_inspector_visible)
        self.shortcut_focus = QtWidgets.QShortcut(QtGui.QKeySequence(
            cfg.variable_inspector_focus_shortcut), self.main_window,
            self.focus, context=QtCore.Qt.ApplicationShortcut)

    def focus(self):
        r"""Makes the dock visible and sets the focus to the filter box."""
        self.set_visible(True)
        self.dock_widget.widget().focus()

    def open_help(self):
        r"""Opens the help tab."""
        self.tabwidget.open_osdoc('manual/variables')

    def set_visible(self, visible):
        r"""Sets the visibility of the dock widget.

        Parameters
        ----------
        visible : bool
        """
        cfg.variable_inspector_visible = visible
        self.set_checked(visible)
        if visible:
            if self.need_refresh:
                self.dock_widget.widget().refresh()
                self.need_refresh = False
            self.dock_widget.show()
            self.dock_widget.widget().focus()
        else:
            self.dock_widget.hide()

    def activate(self):
        r"""Toggles the visibility of the dock widget."""
        self.set_visible(not cfg.variable_inspector_visible)

    def refresh(self):
        r"""Refreshes the variable inspector."""
        if self.dock_widget.isVisible():
            self.dock_widget.widget().refresh()
            self.need_refresh = False
        else:
            self.need_refresh = True

    # The following events all refresh the variable inspector

    def event_heartbeat(self):
        self.refresh()

    def event_change_item(self, name):
        self.refresh()

    def event_pause_experiment(self):
        self.refresh()

    def event_set_workspace_globals(self, global_dict):
        self.dock_widget.widget().set_workspace_globals(global_dict)
        self.refresh()

    def event_open_experiment(self, path):
        self.dock_widget.widget().set_workspace_globals({})
        self.refresh()



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from openexp import resources
from libopensesame import misc
from libqtopensesame.widgets.base_widget import BaseWidget
from libqtopensesame.misc import theme
from libqtopensesame.misc.config import cfg
from libopensesame.oslogging import oslogger
from libqtopensesame.runners import RUNNER_LIST, DEFAULT_RUNNER
from qtpy import QtCore, QtWidgets
from qtpy.QtGui import QIcon
import os


class PreferencesWidget(BaseWidget):

    r"""The widget for the preferences tab."""
    def __init__(self, main_window):
        r"""Constructor.

        Parameters
        ----------
        main_window
            A qtopensesame object.
        """
        super().__init__(main_window, ui='widgets.preferences_widget')
        self.tab_name = '__preferences__'
        self.lock = False
        # Connect the controls
        self.ui.button_experiment_properties.clicked.connect(
            self.tabwidget.open_general)
        self.ui.checkbox_toolbar_text.toggled.connect(self.apply)
        self.ui.checkbox_small_toolbar.toggled.connect(self.apply)
        self.ui.combobox_runner.currentIndexChanged.connect(self.apply)
        self.ui.combobox_style.currentIndexChanged.connect(self.apply)
        self.ui.combobox_theme.currentIndexChanged.connect(self.apply)
        self.ui.combobox_locale.currentIndexChanged.connect(self.apply)
        self.ui.edit_search_plugins.textChanged.connect(self._refresh_plugins)
        
        self.set_controls()
        for ext in self.extensions:
            try:
                w = ext.settings_widget()
            except Exception as e:
                self.notify(f'Extension {ext.name} failed to return settings '
                            f'widget (see debug window for stack trace)',
                            category='warning')
                self.console.write(e)
                continue
            if w is None:
                continue
            if hasattr(w, '__advanced__'):
                self.ui.layout_advanced.addRow(w, None)
            else:
                self.ui.layout_preferences.addRow(w, None)
        self.ui.container_widget.adjustSize()
        self._refresh_plugins()
            
    def _refresh_plugins(self):
        while self.ui.layout_plugins.count():
            child = self.ui.layout_plugins.takeAt(0)
            if child.widget():
                child.widget().deleteLater()
        needle = self.ui.edit_search_plugins.text().strip()
        plugins = list(self.plugin_manager) + \
            list(self.unloaded_extension_manager)
        for plugin in sorted(plugins, key=lambda plugin: plugin.name.lower()):
            if needle and not (needle.lower() in plugin.name.lower() or 
                               needle.lower() in plugin.description.lower()):
                continue
            widget = PluginInfoWidget(self.main_window, plugin)
            self.ui.layout_plugins.addWidget(widget)
        self.ui.layout_plugins.addStretch()

    def set_controls(self):
        r"""Updates the controls."""
        if self.lock:
            return
        self.lock = True
        self.ui.checkbox_toolbar_text.setChecked(
            self.main_window.ui.toolbar_main.toolButtonStyle() ==
            QtCore.Qt.ToolButtonTextUnderIcon
        )
        self.ui.checkbox_small_toolbar.setChecked(cfg.toolbar_size == 16)
        try:
            runner_index = RUNNER_LIST.index(cfg.runner)
        except ValueError:
            oslogger.warning('invalid runner: {}'.format(cfg.runner))
            runner_index = RUNNER_LIST.index(DEFAULT_RUNNER)
        self.ui.combobox_runner.setCurrentIndex(runner_index)
        # Set the locale combobox
        self.ui.combobox_locale.addItem('[Default]')
        self.ui.combobox_locale.setCurrentIndex(0)
        locales = ['en (English)']
        for locale in os.listdir(resources['locale']):
            if locale == 'translatables.qm':
                continue
            locale = locale[:-3]
            locales.append(
                f'{locale} ({QtCore.QLocale(locale).nativeLanguageName()})')
        for i, locale in enumerate(sorted(locales)):
            self.ui.combobox_locale.addItem(locale)
            if cfg.locale == locale[:2]:
                self.ui.combobox_locale.setCurrentIndex(i + 1)
        # Set the style combobox
        self.ui.combobox_style.addItem("[Default]")
        self.ui.combobox_style.setCurrentIndex(0)
        for i, style in enumerate(QtWidgets.QStyleFactory.keys()):
            self.ui.combobox_style.addItem(style)
            if cfg.style == style:
                self.ui.combobox_style.setCurrentIndex(i + 1)
        # Set the theme combobox
        for i, _theme in enumerate(theme.available_themes):
            self.ui.combobox_theme.addItem(_theme)
            if cfg.theme == _theme:
                self.ui.combobox_theme.setCurrentIndex(i)
        self.lock = False
        
    def apply(self):
        """Apply the controls"""
        if self.lock:
            return
        self.lock = True
        self.main_window.ui.toolbar_main.setToolButtonStyle(
            QtCore.Qt.ToolButtonTextUnderIcon
            if self.ui.checkbox_toolbar_text.isChecked() else
            QtCore.Qt.ToolButtonIconOnly
        )
        # Apply locale
        cfg.locale = self.ui.combobox_locale.currentText()[:2]
        if cfg.locale == '[De':  # default
            cfg.locale = ''
        # Apply toolbar size
        old_size = cfg.toolbar_size
        new_size = 16 if self.ui.checkbox_small_toolbar.isChecked() else 32
        if old_size != new_size:
            cfg.toolbar_size = new_size
            self.theme.set_toolbar_size(cfg.toolbar_size)
        # Apply runner
        cfg.runner = RUNNER_LIST[self.ui.combobox_runner.currentIndex()]
        # Apply theme and style
        cfg.theme = self.ui.combobox_theme.currentText()
        cfg.style = self.ui.combobox_style.currentText()
        self.main_window.save_state()
        self.lock = False
        

class PluginInfoWidget(BaseWidget):
    """A widget that holds plugin and extension information for the preferences
    widget. It also contains a button to enabled or disabled the plugin.
    
    Parameters
    ----------
    main_window
    plugin: Plugin or OldStylePlugin
    """

    def __init__(self, main_window, plugin):
        super().__init__(main_window, ui='widgets.plugin_info')
        self._plugin = plugin
        self._cfg_key = f'plugin_enabled_{plugin.name}'
        if self._cfg_key not in cfg:
            cfg.register(self._cfg_key, True)
        self._toggle_enabled(cfg[self._cfg_key])
        self.ui.title.setText(plugin.name)
        self.ui.icon.setPixmap(self.theme.qpixmap(plugin.icon, size=16))
        self.ui.toggle.clicked.connect(self._toggle_enabled)
        self.ui.description.setText(plugin.description)
        if plugin.attribute('__version__', False):
            self.ui.version.setText(plugin['__version__'])
        else:
            self.ui.label_version.setVisible(False)
            self.ui.version.setVisible(False)
        if plugin.attribute('author', False):
            self.ui.author.setText(plugin['author'])
        else:
            self.ui.label_author.setVisible(False)
            self.ui.author.setVisible(False)
        self.ui.folder.setText(plugin.folder)
        self.adjustSize()
        
    def _toggle_enabled(self, enabled):
        self.ui.toggle.setChecked(enabled)
        cfg[self._cfg_key] = enabled
        if enabled:
            self.ui.toggle.setText(_('Enabled'))
            self.ui.toggle.setIcon(self.theme.qicon('user-available'))
            return
        self.ui.toggle.setText(_('Disabled'))
        self.ui.toggle.setIcon(self.theme.qicon('user-away'))
        

# Alias for backwards compatibility
preferences_widget = PreferencesWidget



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from qtpy import QtCore, QtWidgets
from libqtopensesame.misc.config import cfg
from libopensesame.oslogging import oslogger
from libqtopensesame.extensions import BaseExtension
from libqtopensesame.misc.translate import translation_context
_ = translation_context(u'toolbar_menu', category=u'extension')


class ToolbarMenu(BaseExtension):

    r"""Integrates the menu into the toolbar."""
    def event_startup(self):

        self._menu = None
        if not cfg.toolbar_menu_active:
            return
        cfg.toolbar_menu_active = False  # It will be toggled!
        self.activate()
        self.set_checked(True)

    def activate(self):
        r"""Toggle the menubar integration."""
        if self._menu is None:
            self._init_menu()
        if cfg.toolbar_menu_active:
            self._deactivate_toolbar_menu()
        else:
            self._activate_toolbar_menu()

    def _init_menu(self):
        r"""Creates a widget with a menu that copies all actions from the main
        menu bar.
        """
        self._menu = QtWidgets.QMenu()
        for action in self.menubar.actions():
            self._menu.addAction(action)
        self.stretch = QtWidgets.QWidget()
        self.stretch.setSizePolicy(
            QtWidgets.QSizePolicy.Expanding,
            QtWidgets.QSizePolicy.Expanding
        )
        self.button = QtWidgets.QPushButton(
            self.theme.qicon(self.icon()),
            _(u'Menu')
        )
        self.button.setMenu(self._menu)
        self.button.setIconSize(
            QtCore.QSize(cfg.toolbar_size, cfg.toolbar_size)
        )
        self.button.setFlat(True)
        self.toolbar.addWidget(self.stretch)
        self.menu_action = self.toolbar.addWidget(self.button)

    def _keep_toolbar_visible(self, visible):

        if visible:
            return
        oslogger.debug('keep toolbar visible')
        self.toolbar.setVisible(True)

    def _activate_toolbar_menu(self):
        r"""Hide the menu bar and show the toolbar widget."""
        cfg.toolbar_menu_active = True
        self.menubar.setVisible(False)
        self.menu_action.setVisible(True)
        self.stretch.setVisible(True)
        if not self.toolbar.isVisible():
            self.toolbar.setVisible(True)
        self.toolbar.visibilityChanged.connect(self._keep_toolbar_visible)

    def _deactivate_toolbar_menu(self):
        r"""Show the menu bar and hide the toolbar widget."""
        cfg.toolbar_menu_active = False
        self.menubar.setVisible(True)
        self.menu_action.setVisible(False)
        self.stretch.setVisible(False)
        self.toolbar.visibilityChanged.disconnect()



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from libqtopensesame.misc import drag_and_drop
from libqtopensesame.misc.base_component import BaseComponent


class BaseDraggable:

    r"""A base class for components that support drag and drop."""
    def set_supported_drop_types(self, types=None):

        if not hasattr(self, u'setAcceptDrops'):
            raise TypeError(u'Object does not support drops')
        if types is None:
            self.setAcceptDrops(False)
            self.supported_drop_types = None
            return
        self.setAcceptDrops(True)
        self.supported_drop_types = types

    def dragEnterEvent(self, e):
        r"""Handles drag-enter events to see if they are supported

        Parameters
        ----------
        e : QDragEnterEvent
            A drag-enter event.
        """
        if not hasattr(self, u'supported_drop_types'):
            e.ignore()
            return
        data = drag_and_drop.receive(e)
        if drag_and_drop.matches(data, self.supported_drop_types):
            e.accept()
        else:
            e.ignore()

    def dropEvent(self, e):
        r"""Handles drop events and accepts them if supported.

        Parameters
        ----------
        e : QDropEvent
            A drop event.
        """
        if not hasattr(self, u'supported_drop_types'):
            e.ignore()
            return
        data = drag_and_drop.receive(e)
        if drag_and_drop.matches(data, self.supported_drop_types):
            e.accept()
            self.accept_drop(data)
        else:
            e.ignore()

    def accept_drop(self, data):
        r"""Is called after a supported drop type. Should be re-implemented.

        Parameters
        ----------
        data : dict
            The drop data.
        """
        pass


# Alias for backwards compatibility
base_draggable = BaseDraggable



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from qtpy import QtWidgets
from libqtopensesame.misc.base_subcomponent import BaseSubcomponent
from libqtopensesame.misc.translate import translation_context
_ = translation_context(u'unused_widget', category=u'core')


class UnusedItemsContextMenu(BaseSubcomponent, QtWidgets.QMenu):

    r"""Provides a context menu for the unused items bin."""
    def __init__(self, main_window, treeitem):
        r"""Constructor.

        Parameters
        ----------
        main_window : qtopensesame
            The main-window object.
        treeitem : tree_item_item
            The tree item.
        """
        super().__init__(main_window)
        self.setup(main_window)
        a = self.addAction(
            self.theme.qicon(u'purge'),
            _(u'Permanently delete unused items'),
            self.purge_unused
        )
        a.setDisabled(not self.experiment.items.unused())

    def purge_unused(self):
        r"""Purges unused items."""
        self.tabwidget.open_unused()
        w = self.tabwidget.get_widget(u'__unused__')
        w.purge_unused()
        w.on_activate()


# Alias for backwards compatibility
unused_items_context_menu = UnusedItemsContextMenu



"""A simple text input form"""

category = "Form"
icon = "os-form_text_input"
help_url = "manual/forms/readymade"
controls = [
    {
        "label": "Form title",
        "tooltip": "Title to appear above the form text",
        "type": "line_edit",
        "var": "form_title"
    },
    {
        "label": "Response variable",
        "tooltip": "The experimental variable to save the response in",
        "type": "line_edit",
        "var": "form_var"
    },
    {
        "label": "Timeout",
        "tooltip": "A response timeout",
        "type": "line_edit",
        "validator": "timeout",
        "var": "timeout"
    },
    {
        "label": "Your question",
        "tooltip": "A question text",
        "type": "editor",
        "var": "form_question"
    }
]



# coding=utf-8

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from openexp._canvas._ellipse.ellipse import Ellipse
from openexp._canvas._element.legacy import LegacyElement
import pygame


class Legacy(LegacyElement, Ellipse):

    def prepare(self):

        x = int(self.x)
        y = int(self.y)
        w = int(self.w)
        h = int(self.h)
        x, y = self.to_xy(x, y)
        if self.fill:
            pygame.draw.ellipse(self.surface, self.color.backend_color,
                                (x, y, w, h), 0)
            return
        # Use experyiment's method to draw ellipses with a transparent
        # interior by using the transparent colorkey method. This involves
        # some trickery by making a separate surface of which a certain color
        # is designated as transparent and draw the ellipse on there first.
        # When being blit onto another surface, this transparent color
        # (e.g. the ellipses interior is not blitted with the rest.
        line_width = self.penwidth
        size = (w, h)
        surface = pygame.surface.Surface(
            [p + line_width for p in size],
            pygame.SRCALPHA).convert_alpha()
        pygame.draw.ellipse(surface, (0, 0, 0), pygame.Rect(
            (0, 0),
            [p + line_width for p in size]))
        tmp = pygame.surface.Surface(
            [p - line_width for p in size])
        tmp.fill([0, 0, 0])
        tmp.set_colorkey([255, 255, 255])
        hole = pygame.surface.Surface(
            [p - line_width for p in size],
            pygame.SRCALPHA).convert_alpha()
        pygame.draw.ellipse(tmp, (255, 255, 255), pygame.Rect(
            (0, 0), [p - line_width for p in size]))
        hole.blit(tmp, (0, 0))
        surface.blit(hole, (line_width, line_width),
                     special_flags=pygame.BLEND_RGBA_MIN)
        surface.fill(self.color.backend_color,
                     special_flags=pygame.BLEND_RGB_MAX)
        # The line_width affects the temp's surface size, so use it to correct the
        # positioning when blitting.
        self.surface.blit(surface, (x-line_width/2, y-line_width/2))



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *


class Clock:
    """The `clock` object offers basic time functions. A `clock` object is
    created automatically when the experiment starts.

    __Example__

    ~~~ .python
    # Get the timestamp before and after sleeping for 1000 ms
    t0 = clock.time()
    clock.sleep(1000)
    t1 = clock.time()
    time_passed = t1 - t0
    print(f'This should be 1000: {time_passed}')
    ~~~

    [TOC]
    """
    def __init__(self, experiment):
        r"""Constructor to create a new `clock` object. You do not generally
        call this constructor directly, because a `clock` object is created
        automatically when the experiment is launched.

        Parameters
        ----------
        experiment : experiment
            The experiment object.
        """
        self._last_tick = None
        self.experiment = experiment

    def time(self):
        r"""Gives a current timestamp in milliseconds. The absolute meaning of
        the timestamp (i.e. when it was 0) depends on the backend.

        Returns
        -------
        float
            A timestamp.

        Examples
        --------
        >>> t = clock.time()
        >>> print(f'The current time is {t}')
        """
        raise NotImplementedError()

    def sleep(self, ms):
        r"""Sleeps (pauses) for a period.

        Parameters
        ----------
        ms : int, float
            The number of milliseconds to sleep for.

        Examples
        --------
        >>> # Create two canvas objects ...
        >>> my_canvas1 = Canvas()
        >>> my_canvas1.text('1')
        >>> my_canvas2 = Canvas()
        >>> my_canvas2.text('2')
        >>> # ... and show them with 1 s in between
        >>> my_canvas1.show()
        >>> clock.sleep(1000)
        >>> my_canvas2.show()
        """
        raise NotImplementedError()

    def loop_for(self, ms, throttle=None, t0=None):
        r"""*New in v3.2.0*

        An iterator that loops for a fixed time.

        Parameters
        ----------
        ms : int. float
            The number of milliseconds to loop for.
        throttle : NoneType, float, int, optional
            A period to sleep for in between each iteration.
        t0 : NoneType, float, int, optional
            A starting time. If `None`, the starting time is the moment at
            which the iteration starts.

        Returns
        -------

        Examples
        --------
        >>> for ms in clock.loop_for(100, throttle=10):
        >>>     print(ms)
        """
        if t0 is None:
            t0 = self.time()
        while True:
            dt = self.time() - t0
            if dt >= ms:
                break
            yield dt
            if throttle is not None:
                self.sleep(throttle)

    def once_in_a_while(self, ms=1000):
        r"""*New in v3.2.0*

        Periodically returns `True`. This is mostly useful
        for executing
        code (e.g. within a `for` loop) that should only be
        executed once
        in a while.

        Parameters
        ----------
        ms : int, float, optional
            The minimum waiting period.

        Returns
        -------
        bool
            `True` after (at least) the minimum waiting period has
            passed since
            the last call to `Clock.once_in_a_while()`, or
            `False` otherwise.

        Examples
        --------
        >>> for i in range(1000000):
        >>>     if clock.once_in_a_while(ms=50):
        >>>         # Execute this code only once every 50 ms
        >>>         print(clock.time())
        """
        now = self.time()
        if self._last_tick is not None and now - self._last_tick < ms:
            return False
        self._last_tick = now
        return True


# Non PEP-8 alias for backwards compatibility
clock = Clock



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
import platform
import pygame
from pygame.locals import *
from string import whitespace
from openexp._keyboard.keyboard import Keyboard
from openexp.backend import configurable

# Whitespace, backspace, and empty strings are not acceptable names for keys.
# These should be converted to descriptions, e.g. '\t' to 'tab'
invalid_unicode = [u'', u'\x08', u'\x7f'] + list(whitespace)
# On mac arrow keys are not accepted as valid input either. Add them to this list
if platform.system() == "Darwin":
    invalid_unicode += [
        u'\uf702',  # left
        u'\uf703',  # right
        u'\uf700',  # up
        u'\uf701',  # down
        u'\uf729',  # home
        u'\uf72b',  # end
        u'\uf72c',  # page up
        u'\uf72d',  # page down
        u'\uf728',  # delete
        u'\uf739',  # numlock
    ]


class Legacy(Keyboard):

    r"""This is a keyboard backend built on top of PyGame. For function
    specifications and docstrings, see `openexp._keyboard.keyboard`.
    """
    def __init__(self, experiment, **resp_args):

        pygame.init()
        self.key_code_to_name = {}
        self.key_name_to_code = {}
        for i in dir(pygame):
            if i[:2] == u"K_":
                code = getattr(pygame, i)
                name1 = self.key_name(code).lower()
                name2 = name1.upper()
                name3 = i[2:].lower()
                name4 = name3.upper()
                self.key_code_to_name[code] = [name1, name2, name3, name4]
                try:
                    i = int(name5)
                    self.key_code_to_name[code].append(name5)
                except:
                    pass
                self.key_name_to_code[name1] = code
                self.key_name_to_code[name2] = code
                self.key_name_to_code[name3] = code
                self.key_name_to_code[name4] = code
        self.persistent_virtual_keyboard = False
        Keyboard.__init__(self, experiment, **resp_args)

    @configurable
    def get_key(self):

        return self._get_key_event(pygame.KEYDOWN)

    @configurable
    def get_key_release(self):

        return self._get_key_event(pygame.KEYUP)

    def _get_key_event(self, event_type):

        start_time = pygame.time.get_ticks()
        time = start_time
        keylist = self.keylist
        timeout = self.timeout
        while True:
            time = pygame.time.get_ticks()
            # Some input methods send multiple key events at the same time,
            # for example when composing a multicharacter Chinese or Japanese
            # string. That's why we process up all events, rather than
            # assuming that there's only a single relevant event in the queue.
            key = u''
            for event in pygame.event.get(event_type):
                if event.key == pygame.K_ESCAPE:
                    self.experiment.pause()
                # KEYUP events don't have a unicode property, so in that case
                # we fall back to converting the key code straight to an ASCII
                # value. This is not great, because it assumes a QWERTY
                # keyboard layout.
                if hasattr(event, u'unicode'):
                    ucode = event.unicode
                elif event.key < 128:
                    ucode = chr(event.key)
                else:
                    ucode = u''
                if ucode in invalid_unicode:
                    key += self.key_name(event.key)
                else:
                    key += ucode
            if key and (keylist is None or key in keylist):
                return key, time
            if timeout is not None and time - start_time >= timeout:
                break
        return None, time

    def get_mods(self):

        l = []
        mods = pygame.key.get_mods()
        if mods & KMOD_LSHIFT or mods & KMOD_RSHIFT or mods & KMOD_SHIFT:
            l.append(u"shift")
        if mods & KMOD_LCTRL or mods & KMOD_RCTRL or mods & KMOD_CTRL:
            l.append(u"ctrl")
        if mods & KMOD_LALT or mods & KMOD_RALT or mods & KMOD_ALT:
            l.append(u"alt")
        if mods & KMOD_LMETA or mods & KMOD_RMETA or mods & KMOD_META:
            l.append(u"meta")
        return l

    def valid_keys(self):

        return sorted(self.key_name_to_code.keys())

    def synonyms(self, key):

        # If the key is not familiar, simply return it plus its string
        # representation.
        if key not in self.key_name_to_code:
            return [key, safe_decode(key)]
        return self.key_code_to_name[self.key_name_to_code[key]]

    def flush(self):

        keypressed = False
        for event in pygame.event.get():
            if event.type == KEYDOWN:
                keypressed = True
                if event.key == pygame.K_ESCAPE:
                    self.experiment.pause()
        pygame.event.pump()
        return keypressed

    def key_name(self, key):

        return str(pygame.key.name(key)).replace(u'[', u'').replace(u']',
                                                                    u'')

    def _keycode_to_str(self, keycode):

        return safedecode(pygame.key.name(keycode)) \
            .replace(u'[', u'').replace(u']', u'').lower()


# Non PEP-8 alias for backwards compatibility
legacy = Legacy



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from qtpy import QtCore, QtWidgets
from libqtopensesame.dialogs.base_dialog import BaseDialog
from libqtopensesame.widgets import SketchpadWidget


class SketchpadDialog(BaseDialog):

    r"""The pop-out version of the sketchpad_widget."""
    def __init__(self, main_window, sketchpad):
        r"""Constructor.

        Parameters
        ----------
        main_window
            A qtopensesame object.
        sketchpad
            A sketchpad object.
        """
        super().__init__(main_window,
                         flags=QtCore.Qt.WindowMinMaxButtonsHint | 
                               QtCore.Qt.WindowCloseButtonHint)
        self.sketchpad = sketchpad
        self.tools_widget = sketchpad_widget.sketchpad_widget(self.sketchpad,
                                                              parent=self,
                                                              embed=False)
        self.close_button = QtWidgets.QPushButton(self.theme.qicon(u"close"),
                                                  u"Close")
        self.close_button.setIconSize(QtCore.QSize(16, 16))
        self.close_button.clicked.connect(self.accept)
        self.hbox = QtWidgets.QHBoxLayout()
        self.hbox.addStretch()
        self.hbox.addWidget(self.close_button)
        self.hbox.setContentsMargins(0, 0, 0, 0)
        self.hbox_widget = QtWidgets.QWidget()
        self.hbox_widget.setLayout(self.hbox)
        self.vbox = QtWidgets.QVBoxLayout()
        self.vbox.addWidget(self.tools_widget)
        self.vbox.addWidget(self.hbox_widget)
        self.setLayout(self.vbox)


# Alias for backwards compatibility
sketchpad_dialog = SketchpadDialog



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
import time
import os
import json
from libopensesame.py3compat import *
from libopensesame.exceptions import UserAborted
from libqtopensesame.extensions import BaseExtension
from libopensesame import misc
from libqtopensesame.misc.translate import translation_context
_ = translation_context(u'after_experiment', category=u'extension')


class AfterExperiment(BaseExtension):

    def event_end_experiment(self, ret_val):
        r"""Handles the end of an experiment.

        Parameters
        ----------
        ret_val : Exception, NoneType
            An Exception, or None if no exception occurred.
        """
        if ret_val is None:
            self.handle_success()
        else:
            self.handle_exception(ret_val)

    @property
    def _logfile(self):
        """Returns the path to the logfile of the last run, or None if this
        could not be determined.
        """
        # Depending on the runner, the logfile is either a property of the
        # var object ...
        var = self.extension_manager.provide(
            'jupyter_workspace_variable',
            name='var'
        )
        try:
            if var and 'logfile' in var:
                return var.logfile
        except TypeError:
            # JupyterConsole returns a custom FailedToGetWorkspaceVariable
            # Exception when an error occurs. This is not iterable and results
            # in a TypeError.
            pass
        # ... or it is a global variable in the workspace
        logfile = self.extension_manager.provide(
            'jupyter_workspace_variable',
            name='logfile'
        )
        if logfile:
            return logfile

    @property
    def _extra_data_files(self):
        """Returns list of extra data files, such as eye-tracking data."""
        data_files = self.extension_manager.provide(
            'jupyter_workspace_variable',
            name='data_files'
        )
        if not data_files:
            return
        try:
            return [path for path in data_files if path != self._logfile]
        except TypeError:
            # JupyterConsole returns a custom FailedToGetWorkspaceVariable
            # Exception when an error occurs. This is not iterable and results
            # in a TypeError.
            pass

    def event_after_experiment_copy_logfile(self):
        r"""Copies the logfile to the file pool."""
        if self._logfile is None:
            return
        self.main_window.ui.pool_widget.add([self._logfile],
                                            rename=True)

    def event_after_experiment_open_logfile_folder(self):
        r"""Opens the logfile folder."""
        if self._logfile is None:
            return
        misc.open_url(os.path.dirname(self._logfile))

    def event_after_experiment_open_logfile(self):
        r"""Opens the logfile."""
        if self._logfile is None:
            return
        misc.open_url(self._logfile)

    def handle_success(self):
        r"""Shows a summary after successful completion of the experiment."""
        logfile = self._logfile
        if logfile is None:
            logfile = u'Unknown logfile'
        md = safe_read(self.ext_resource(u'finished.md')) % {
            u'time': time.ctime(),
            u'logfile': logfile
        }
        if self._extra_data_files:
            md += u'\n' \
                + _(u'The following extra data files where created:') + u'\n\n'
            for data_file in self._extra_data_files:
                md += u'- `' + data_file + u'`\n'
            md += u'\n'
        self.tabwidget.open_markdown(
            md, u'os-finished-success', _(u'Finished'))

    def handle_exception(self, e):
        r"""Shows a summary when the experiment was aborted.

        Parameters
        ----------
        e : Exception
            The Exception that caused the experiment to stop.
        """
        if isinstance(e, UserAborted):
            self.extension_manager.fire('notify', message=str(e),
                                        always_show=True)
            return
        self.console.write(e)
        self.tabwidget.open_markdown(e.markdown(), 'os-finished-error',
                                     e.title())



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from libopensesame.oslogging import oslogger
from qtpy import QtWidgets
from libqtopensesame.widgets.general_header_widget import GeneralHeaderWidget
from libqtopensesame.widgets.base_widget import BaseWidget
from libopensesame.exceptions import InvalidColor, MissingDependency
from openexp._color.color import Color
from openexp import backend
from libqtopensesame.misc.translate import translation_context
_ = translation_context(u'general_properties', category=u'core')


class GeneralProperties(BaseWidget):
    """The QWidget for the general properties tab.

    Parameters
    ----------
    main_window
        A qtopensesame object.
    """

    def __init__(self, main_window):
        super().__init__(main_window, ui=u'widgets.general_properties')
        self.lock = False
        # Set the header, with the icon, label and script button
        self.header_widget = GeneralHeaderWidget(self, self.main_window)
        header_hbox = QtWidgets.QHBoxLayout()
        header_hbox.addWidget(self.theme.qlabel(u"experiment"))
        header_hbox.addWidget(self.header_widget)
        header_hbox.addStretch()
        header_hbox.setContentsMargins(0, 0, 0, 0)
        header_hbox.setSpacing(12)
        header_widget = QtWidgets.QWidget()
        header_widget.setLayout(header_hbox)
        self.ui.container_layout.insertWidget(0, header_widget)
        # Initialize the color and font widgets
        self.ui.edit_foreground.initialize(self.experiment)
        self.ui.edit_background.initialize(self.experiment)
        self.ui.widget_font.initialize(self.experiment)
        self.ui.button_reset_backend.clicked.connect(self._reset_backend)
        # Set the backend combobox
        self._backend_button_group = QtWidgets.QButtonGroup()
        for id_, (name, info) in enumerate(backend.backend_info().items()):
            radio_button = QtWidgets.QRadioButton(
                f'{info["description"]} ({name})')
            self._backend_button_group.addButton(radio_button)
            self._backend_button_group.setId(radio_button, id_)
            self.ui.layout_backend_list.addWidget(radio_button)
        self.quick_connect(
            slot=self.main_window.ui.tabwidget.open_general_script,
            signals=[self.ui.button_script_editor.clicked],
        )
        self.quick_connect(
            slot=self.main_window.ui.tabwidget.open_backend_settings,
            signals=[self.ui.button_backend_settings.clicked]
        )
        self.quick_connect(
            slot=self.apply_changes,
            signals=[
                self.ui.spinbox_width.editingFinished,
                self.ui.spinbox_height.editingFinished,
                self.ui.checkbox_disable_garbage_collection.stateChanged,
                self.ui.edit_foreground.textEdited,
                self.ui.edit_background.textEdited,
                self.ui.widget_font.font_changed,
                self._backend_button_group.idClicked,
            ])
        self.tab_name = u'__general_properties__'
        self.on_activate = self.refresh

    def set_header_label(self):
        """Sets the general header based on the experiment title and
        description.
        """
        self.header_widget.set_name(self.experiment.var.title)
        self.header_widget.set_desc(self.experiment.var.description)

    def apply_changes(self):
        """Applies changes to the general tab."""
        # Skip if the general tab is locked and lock it otherwise
        if self.lock:
            return
        self.lock = True

        self.main_window.set_busy(True)
        self.main_window.extension_manager.fire(u'prepare_change_experiment')
        # Set the title and the description
        title = self.experiment.syntax.sanitize(
            self.header_widget.edit_name.text())
        if title != self.experiment.var.title:
            self.experiment.var.title = title
            self.experiment.build_item_tree()
        desc = self.experiment.syntax.sanitize(
            self.header_widget.edit_desc.text())
        self.experiment.var.description = desc
        # Set the backend
        if self.ui.widget_backend_list.isEnabled():
            i = self._backend_button_group.checkedId()
            name, info = list(backend.backend_info().items())[i]
            oslogger.info(f'setting backend to {name}')
            self.experiment.var.canvas_backend = info["canvas"]
            self.experiment.var.keyboard_backend = info["keyboard"]
            self.experiment.var.mouse_backend = info["mouse"]
            self.experiment.var.sampler_backend = info["sampler"]
            self.experiment.var.clock_backend = info["clock"]
            self.experiment.var.color_backend = info["color"]
            self.ui.button_backend_settings.setEnabled(info['settings'])
        else:
            oslogger.debug(
                u'not setting back-end, because a custom backend is selected')
            self.ui.button_backend_settings.setEnabled(False)
        # Set the display width
        width = self.ui.spinbox_width.value()
        height = self.ui.spinbox_height.value()
        if self.experiment.var.width != width or \
                self.experiment.var.height != height:
            self.main_window.update_resolution(width, height)
        # Set the foreground color. If there are no variables in the color
        # definition, then we check whether it is a valid color specification.
        # If not, then we revert the change and notify the user.
        foreground = self.experiment.syntax.sanitize(
            self.ui.edit_foreground.text())
        if not self.experiment.syntax.contains_variables(foreground):
            try:
                Color.to_hex(foreground)
            except (InvalidColor, MissingDependency) as e:
                self.notify(f'Invalid color: {e}')
                foreground = self.experiment.var.foreground
        self.ui.edit_foreground.setText(foreground)
        self.experiment.var.foreground = foreground
        # Set the background color
        background = self.experiment.syntax.sanitize(
            self.ui.edit_background.text())
        if not self.experiment.syntax.contains_variables(background):
            try:
                Color.to_hex(background)
            except (InvalidColor, MissingDependency) as e:
                self.notify(f'Invalid color: {e}')
                background = self.experiment.var.background
        self.ui.edit_background.setText(background)
        self.experiment.var.background = background
        # Set the font
        self.experiment.var.font_family = self.ui.widget_font.family
        self.experiment.var.font_size = self.ui.widget_font.size
        self.experiment.var.font_italic = self.ui.widget_font.italic
        self.experiment.var.font_bold = self.ui.widget_font.bold
        # Other checkboxes
        self.experiment.var.disable_garbage_collection = \
            self.ui.checkbox_disable_garbage_collection.isChecked()
        # Refresh the interface and unlock the general tab
        self.lock = False
        self.main_window.extension_manager.fire(u'change_experiment')
        self.main_window.set_busy(False)
        self.main_window.set_unsaved()
        
    def _reset_backend(self):
        """Resets the backend to the default backend settings"""
        self.experiment.var.canvas_backend = 'psycho'
        del self.experiment.var.keyboard_backend
        del self.experiment.var.mouse_backend
        del self.experiment.var.sampler_backend
        del self.experiment.var.synth_backend
        del self.experiment.var.color_backend
        del self.experiment.var.clock_backend
        del self.experiment.var.log_backend
        self.refresh()

    def refresh(self):
        """Updates the controls of the general tab."""
        # Lock the general tab to prevent a recursive loop
        self.lock = True
        # Set the header containing the title etc
        self.set_header_label()
        # Select the backend
        backend_name = backend.backend_match(self.experiment)
        if backend_name == u"custom":
            self.ui.widget_reset_backend.setVisible(True)
            self.ui.widget_backend_list.setDisabled(True)
            self.ui.button_backend_settings.setDisabled(True)
        else:
            self.ui.widget_reset_backend.setVisible(False)
            backend_info = backend.backend_info()[backend_name]
            backend_desc = backend_info['description']
            backend_settings = backend_info['settings']
            self.ui.button_backend_settings.setEnabled(backend_settings)
            self.ui.widget_backend_list.setDisabled(False)
            for radio_button in self._backend_button_group.buttons():
                if radio_button.text().startswith(backend_desc):
                    radio_button.setChecked(True)
        # Set the resolution
        try:
            self.ui.spinbox_width.setValue(int(self.experiment.var.width))
            self.ui.spinbox_height.setValue(int(self.experiment.var.height))
        except ValueError:
            self.experiment.notify(
                _(u"Failed to parse the resolution. Expecting positive numeric values.")
            )
        # Set the colors
        self.ui.edit_foreground.setText(safe_decode(
            self.experiment.var.foreground))
        self.ui.edit_background.setText(safe_decode(
            self.experiment.var.background))
        self.ui.widget_font.initialize(self.experiment)
        self.ui.checkbox_disable_garbage_collection.setChecked(
            self.experiment.var.disable_garbage_collection == u'yes')
        # Release the general tab
        self.lock = False


# Alias for backwards compatibility
general_properties = GeneralProperties



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from openexp import resources
from openexp.canvas_elements import Rect


class Plain:

    def __init__(self, form):

        self.form = form
        if not self.form.clicks:
            return
        from openexp.sampler import Sampler
        import os
        self.click_sound = Sampler(self.form.experiment,
                                   resource['widgets/interaction.ogg'],
                                   duration=0, volume=.5)

    @property
    def canvas(self):
        return self.form.canvas

    def box(self, x, y, checked=False):

        return Rect(x, y, 16, 16, fill=checked).construct(self.canvas)

    def box_size(self):

        return 16

    def frame(self, x, y, w, h, style='normal'):

        return Rect(x, y, w, h).construct(self.canvas)

    def click(self):

        if self.form.clicks:
            self.click_sound.play()


# Alias for backwards compatibility
plain = Plain



"""A Jupyter console"""

label = "Show console"
tooltip = "Show Jupyter/ IPython console"
priority = 101
icon = "os-debug"
settings = {
    "jupyter_inprocess": False,
    "jupyter_visible": True,
    "jupyter_focus_shortcut": "Ctrl+3",
    "jupyter_kernels": {
        "ide": "",
        "default": "python"
    },
    "jupyter_default_kernel": "python"
}
shortcut = "Ctrl+D"
checkable = True
menu = {
    "index": 8,
    "submenu": "View"
}
toolbar = {
    "index": 13
}
modes = ["default", "ide"]



# -*- coding:utf-8 -*-

"""
This file is part of openexp.

openexp is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

openexp is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with openexp.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from libqtopensesame.sketchpad_elements._arrow import arrow
from libqtopensesame.sketchpad_elements._circle import circle
from libqtopensesame.sketchpad_elements._ellipse import ellipse
from libqtopensesame.sketchpad_elements._fixdot import fixdot
from libqtopensesame.sketchpad_elements._gabor import gabor
from libqtopensesame.sketchpad_elements._image import image
from libqtopensesame.sketchpad_elements._line import line
from libqtopensesame.sketchpad_elements._noise import noise
from libqtopensesame.sketchpad_elements._textline import textline
from libqtopensesame.sketchpad_elements._rect import rect
elements = [textline, image, fixdot, line, arrow, rect, circle, ellipse, gabor,
            noise]



# -*- coding:utf-8 -*-

"""
This file is part of openexp.

openexp is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

openexp is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with openexp.  If not, see <http://www.gnu.org/licenses/>.

---
desc:
    Drag-and-drop data is communicated by json-encoded dictionaries. The
    following drag types exist:

    - type: invalid
    - type: sketchpad-element-move
    - type: url-local
      url: [url]
    - type: item-existing
      item-name: [item name]
      QTreeWidgetItem [representation of QTreeWidgetItem]
    - type: item-clone
      item: [item name]
    - type: variable
      variable: [variable name]
    - type: item-snippet
      main-item-name: [suggested name]
      items:
      - item-name: [suggested name]
        item-type: [item type]
        script: [script]
      - [...]
---
"""
import json
from qtpy import QtCore, QtGui
from libopensesame.oslogging import oslogger
from libopensesame.py3compat import *


# For unknown mimetypes
INVALID_DATA = {
    u'type': u'invalid'
}


def matches(data, types):
    r"""Checks whether a data dictionary matches any of the specified drop
    types.

    Parameters
    ----------
    data    A data dictionary. Non-dict types do not give an error, but return
        False.
    types : list
        A list of types, matching the 'type' key in the data dict.

    Returns
    -------
    bool
        True if the data matches, False otherwise.
    """
    if not isinstance(data, dict):
        return False
    if u'type' not in data:
        return False
    if data[u'type'] not in types:
        return False
    return True


def receive(drop_event):
    r"""Extracts data from a drop event. This data should be embedded in the
    mimedata as a json text string.

    Parameters
    ----------
    drop_event : QDropEvent
        A drop event.

    Returns
    -------
    dict
        A data dictionary. The 'type' key identifies the type of data that is
        being dropped.
    """
    mimedata = drop_event.mimeData()
    # Reject unknown mimedata
    if not mimedata.hasText() and not mimedata.hasUrls():
        oslogger.warning(u'No text data')
        return INVALID_DATA
    # Process url mimedata
    if mimedata.hasUrls():
        urls = mimedata.urls()
        if len(urls) == 0:
            return INVALID_DATA
        url = urls[0]
        if not url.isLocalFile():
            return INVALID_DATA
        path = str(url.toLocalFile())
        data = {
            u'type': u'url-local',
            u'url': path
        }
        return data
    # Process JSON text data
    text = str(mimedata.text())
    try:
        data = json.loads(text)
    except:
        oslogger.warning(u'Failed to load json mime data')
        return INVALID_DATA
    if not isinstance(data, dict) or u'type' not in data:
        return INVALID_DATA
    return data


def send(drag_src, data):
    r"""Starts a drag event, and embeds a data dictionary as json text in the
    mimedata.

    Parameters
    ----------
    drag_src : QWidget
        The source widget.
    data : dict or str
        The data to be sent. If this is a dict, then it is converted to str
        in json format. Otherwise the data is sent as is.

    Returns
    -------
    QDrag
        A drag object. The start function is called automatically.
    """
    text = safe_decode(json.dumps(data) if isinstance(data, dict) else data,
                       enc='utf-8')
    mimedata = QtCore.QMimeData()
    mimedata.setText(text)
    drag = QtGui.QDrag(drag_src)
    drag.setMimeData(mimedata)
    drag.exec_()



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from openexp import backend


def Sampler(experiment, *arglist, **kwdict):
    r"""A factory that returns a back-end specific sampler object.

    Parameters
    ----------
    experiment : experiment
        The experiment object.
    *arglist : list
        See sampler.__init__().
    **kwdict : dict
        See sampler.__init__().
    """
    cls = backend.get_backend_class(experiment, u'sampler')
    return cls(experiment, *arglist, **kwdict)


def init_sound(experiment):
    r"""Calls the back-end specific init_sound function.

    Parameters
    ----------
    experiment
        The experiment object.
    type
        experiment
    """
    cls = backend.get_backend_class(experiment, u'sampler')
    cls.init_sound(experiment)


def close_sound(experiment):
    r"""Calls the back-end specific close_sound function.

    Parameters
    ----------
    experiment
        The experiment object.
    type
        experiment
    """
    cls = backend.get_backend_class(experiment, u'sampler')
    cls.close_sound(experiment)


# Non PEP-8 alias for backwards compatibility
sampler = Sampler



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
import os
import pkgutil
import pathlib
from collections import OrderedDict
from openexp import resources
from importlib import import_module
from libopensesame import plugins  # deprecated
from libopensesame.misc import camel_case, snake_case
from libopensesame.oslogging import oslogger


class Plugin:
    """An unloaded plugin or extension. This is created for each available
    plugin or extension. An instance of the plugin or extension is created by
    calling Plugin.build().

    Attributes defined in the __init__.py of the plugin are available through
    this class using dict syntax Plugin['icon'] and the `in` operator.
    `Plugin.attribute()` allows you to specify a default value for the
    attribute.

    Parameters
    ----------
    mod: module
        The module that contains the plugin
    """
    
    def __init__(self, mod):
        self.name = mod.__package__.split('.')[-1]
        self.icon = 'applications-utilities'
        self._mod = mod
        self._cls = None
        self._type = 'plugins' if \
            self._mod.__package__.startswith('opensesame_plugins') else \
            'extensions'
        self.folder = os.path.dirname(mod.__file__)
            
    def __contains__(self, attr):
        return attr in self._mod.__dict__

    def __getitem__(self, attr):
        return self._mod.__dict__[attr]
        
    def attribute(self, attr, default=None):
        return self._mod.__dict__.get(attr, default)
    
    def build(self, *args, **kwargs):
        if self._cls is None:
            resources.add_resource_folder(self.folder)
            oslogger.debug(f'finding plugin runtime for {self.name}')
            mod = import_module(
                f'{self._mod.__package__}.{self.name}')
            self._cls = self._get_cls(mod)
            if not hasattr(self._cls, 'description'):
                self._cls.description = self.description
        oslogger.debug(f'building plugin gui for {self.name}')
        return self._cls(*args, **kwargs)
        
    @property
    def description(self):
        return self._mod.__doc__
        
    def _get_cls(self, mod):
        
        if hasattr(mod, camel_case(self.name)):
            return getattr(mod, camel_case(self.name))
        return getattr(mod, self.name)


class OldStylePlugin:
    """An adapter that maps the new plugin API onto the old API (<= 3.3). To
    maintain backwards compatibility with old plugins. This is deprecated and
    will be removed in future versions.
    """
    
    def __init__(self, name, type_):
        self.name = name
        self.type_ = type_
        self.icon = 'applications-utilities'
        self.folder = plugins.plugin_folder(name, _type=self.type_)
        
    def __getitem__(self, attr):
        return self.attribute(attr, default=None)
        
    def attribute(self, attr, default=None):
        return plugins.plugin_property(self.name, attr, default=default, 
                                       _type=self.type_)
        
    def build(self, *args, **kwargs):
        oslogger.debug(f'building old-style plugin for {self.name}')
        if self.type_ == 'plugins':
            return plugins.load_plugin(self.name, *args, **kwargs)
        return plugins.load_extension(self.name, *args, **kwargs)
        
    @property
    def description(self):
        return self.attribute('description', 'No description')


class PluginManager:
    """A manager for unloaded plugins and extensions. This scans all available
    plugins from a package (`pkg`) and provides access to these as Plugin
    objects through a dict interface. `PluginManager.filter()` can be used
    to iterate only through plugins that match on specific attributes.

    Parameters
    ----------
    pkg: module
        A plugin or extension module, typically the result of
        `import opensesame_extensions` or `import opensesame_plugins
    """
    
    # These class attributes define which classes should be instantiated for
    # the plugins
    plugin_cls = Plugin
    oldstyle_plugin_cls = OldStylePlugin
    
    def __init__(self, pkg):
        self._plugins = OrderedDict()
        self._pkg = pkg
        self._aliases = {}
        self.sub_packages = []
        for importer, name, ispkg in pkgutil.iter_modules(
                pkg.__path__, prefix=pkg.__name__ + '.'):
            if not ispkg:
                continue
            oslogger.debug(f'found plugin package {name} in {importer.path}')
            self._discover_subpkg(name)
        self._discover_oldstyle()
        # Sort all plugins by their priority, such that high priority values
        # come first
        self._plugins = OrderedDict(
            sorted(self._plugins.items(),
                   key=lambda plugin: -plugin[1].attribute('priority', 0)))
        
    def _discover_subpkg(self, name):
        pkg = import_module(name)
        self.sub_packages.append(pkg)
        for importer, plugin_name, ispkg in pkgutil.iter_modules(
                pkg.__path__, prefix=name + '.'):
            if not ispkg:
                continue
            oslogger.debug(f'found plugin {plugin_name} in {importer.path}')
            self._discover_plugin(plugin_name)
            
    def _discover_plugin(self, name):
        plugin = self.plugin_cls(import_module(name))
        if plugin.name in self._aliases:
            oslogger.warning(
                f'duplicate plugin: {plugin.name} at {plugin.folder} '
                f'already found at '
                f'{self._plugins[self._aliases[plugin.name]].folder}')
            return
        self._register(plugin)
        
    def _discover_oldstyle(self):
        type_ = 'plugins' if self._pkg.__name__ == 'opensesame_plugins' \
            else 'extensions'
        for plugin_name in plugins.list_plugins(_type=type_):
            oslogger.warning(f'found deprecated old-style plugin '
                             f'{plugin_name} in '
                             f'{plugins.plugin_folder(plugin_name, _type=type_)}')
            plugin = self.oldstyle_plugin_cls(plugin_name, type_)
            if plugin.name in self._aliases:
                oslogger.warning(
                    f'duplicate plugin: {plugin.name} at {plugin.folder} '
                    f'already found at '
                    f'{self._plugins[self._aliases[plugin.name]].folder}')
                continue
            self._register(plugin)

    def _register(self, plugin):
        """Registers a plugin and also stores various aliases to deal with
        irregular naming.
        """
        self._plugins[plugin.name] = plugin
        # We remember the plugin under various aliases, which are either
        # specified as a plugin attribute, or derived by turning the name into
        # snake_case or CamelCase if it wasn't already.
        self._aliases[plugin.name] = plugin.name
        for alias in plugin.attribute('aliases', []):
            self._aliases[alias] = plugin.name
        if plugin.name.islower():
            self._aliases[camel_case(plugin.name)] = plugin.name
        else:
            self._aliases[snake_case(plugin.name)] = plugin.name
        
    def filter(self, **kwargs):
        for plugin in self:
            for key, value in kwargs.items():
                attr = plugin.attribute(
                    key, default='default' if key == 'modes' else None)
                if isinstance(attr, (list, tuple, set, dict)):
                    if value in attr:
                        yield plugin
                elif attr == value:
                    yield plugin
        
    def __contains__(self, name):
        return name in self._aliases and self._aliases[name] in self._plugins
    
    def __getitem__(self, name):
        return self._plugins[self._aliases[name]]
        
    def __iter__(self):
        for plugin in self._plugins.values():
            yield plugin



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from libqtopensesame.extensions import BaseExtension
from libqtopensesame.misc.config import cfg
from libopensesame import metadata
from qtpy import QtCore
import sys
import yaml
import os
import traceback
from libqtopensesame.misc.translate import translation_context
_ = translation_context(u'bug_report', category=u'extension')


class BugReport(BaseExtension):

    def event_startup(self):
        sys.excepthook = self.captured_err

    def indent(self, s):
        r"""Tab-indent a piece of text so that it's code for Markdown.

        Parameters
        ----------
        s
            The text to indent.

        Returns
        -------
        Indented text.
        """
        return u'\t' + s.replace(u'\n', u'\n\t').replace(os.linesep, u'\n\t')

    def event_bug_report_send(self):
        r"""Sends a bug report for the latest stacktrace. Also closes the
        current tab, which is the report tab, and shows a results tab.
        """
        self.main_window.tabwidget.close_current()
        from urllib.request import urlopen
        from urllib.parse import urlencode
        if self.traceback is None:
            return
        q = urlencode({
            u'traceback': safe_str(self.traceback, errors=u'ignore'),
            u'version': safe_str(metadata.__version__, errors=u'ignore'),
            u'python_version': safe_str(metadata.python_version,
                                        errors=u'ignore'),
            u'platform': safe_str(metadata.platform, errors=u'ignore'),
        })
        url = cfg.bug_report_url + u'?' + q
        try:
            fd = urlopen(url)
            resp = safe_decode(fd.read(), errors=u'ignore')
            fd.close()
        except:
            self.tabwidget.open_markdown(self.ext_resource(u'failure.md'),
                                         title=_(u'Bug report not sent'))
            return
        if resp == u'OK':
            self.tabwidget.open_markdown(self.ext_resource(u'success.md'),
                                         title=_(u'Bug report sent'))
        else:
            self.tabwidget.open_markdown(self.ext_resource(u'failure.md'),
                                         title=_(u'Bug report not sent'))

    def captured_err(self, exception_type, value, tb):
        r"""Shows a report tab when an error message has been captured."""
        error_list = traceback.format_exception(exception_type, value, tb)
        self.traceback = u"".join([safe_decode(tb_line)
                                  for tb_line in error_list])
        self.traceback_md = u'~~~ .traceback\n%s\n~~~\n' % self.traceback
        md = safe_read(self.ext_resource(u'report.md')) % {
            # u'traceback' : self.indent(self.stderr.buffer),
            u'traceback': self.traceback_md,
            u'version': metadata.__version__,
            u'python_version': safe_str(metadata.python_version,
                                        errors=u'ignore'),
            u'platform': metadata.platform,
        }
        self.tabwidget.open_markdown(md, title=_(u'Oops ...'))
        sys.stderr.write(self.traceback)



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
import os
import sys
import traceback
from qtpy import QtWidgets
from libqtopensesame.misc.config import cfg
from libopensesame.oslogging import oslogger
from libopensesame.experiment import Experiment
from libopensesame.py3compat import *
from libqtopensesame.misc.translate import translation_context
_ = translation_context('base_runner', category='core')


class BaseRunner:
    """
    A runner implements a specific way to execute an OpenSesame experiment from
    within the GUI. The base_runner is an abstract runner that is inherited by
    actual runners.

    Parameters
    ----------
    main_window : QtOpenSesame
    """
    
    valid_logfile_extensions = '.csv', '.txt', '.dat', '.log', '.json'
    supports_kill = False

    def __init__(self, main_window):
        self.main_window = main_window
        self.paused = False

    @property
    def console(self):
        return self.main_window.console

    @property
    def tabwidget(self):
        return self.main_window.tabwidget

    def execute(self):
        """
        Executes the experiments. This function should be transparent and leave
        no mess to clean up for the GUI.

        Returns:
        None if the experiment finished cleanly, or an Exception (any kind) if
        an exception occurred.
        """
        pass

    def get_logfile(self, quick=False, subject_nr=0, file_extension='.csv'):
        """
        Gets the logfile for the current session, either by falling back to a
        default value ('quickrun.csv') or through a pop-up dialogue.

        Parameters
        ----------
        quick : bool, optional
            A boolean to indicate whether default should be used for the
            log-file and subject number. Mostly useful while testing the
            experiment.
        subject_nr : int, optional
        file_extension : str, optional

        Returns
        -------
        str or None
            A pathname for the logfile or None if no logfile was chosen (i.e.
            the dialogue was cancelled).
        """
        remember_logfile = True
        if quick:
            logfile = os.path.join(
                cfg.default_logfile_folder,
                cfg.quick_run_logfile)
            try:
                open(logfile, 'w').close()
                os.remove(logfile)
            except Exception:
                import tempfile
                from libopensesame import misc
                oslogger.warning('failed to open %s' % logfile)
                logfile = os.path.join(
                    safe_decode(
                        tempfile.gettempdir(),
                        enc=sys.getfilesystemencoding()),
                    safe_decode(
                        tempfile.gettempprefix(),
                        enc=sys.getfilesystemencoding()
                    ) + f'quickrun{file_extension}')
                oslogger.warning('Using temporary file %s' % logfile)
                remember_logfile = False
        else:
            # Suggested filename
            suggested_path = os.path.join(
                cfg.default_logfile_folder,
                f'subject-{subject_nr}{file_extension}'
            )
            # Get the data file
            file_filter = f'Log file ({file_extension})'
            logfile = QtWidgets.QFileDialog.getSaveFileName(
                self.main_window.ui.centralwidget,
                _("Choose location for logfile (press 'escape' for default location)"),
                suggested_path, filter=file_filter)
            # In PyQt5, the QFileDialog.getOpenFileName returns a tuple instead
            # of a string, of which the first position contains the path.
            if isinstance(logfile, tuple):
                logfile = logfile[0]
            # An empty string indicates that the dialogue was cancelled, in
            # which case we fall back to a default location.
            if logfile == '':
                logfile = os.path.join(cfg.default_logfile_folder,
                                       f'defaultlog{file_extension}')
            # If a logfile was provided, but it did not have a proper
            # extension, we add a `.csv` extension.
            else:
                if os.path.splitext(logfile)[1].lower() not in \
                        self.valid_logfile_extensions:
                    logfile += file_extension
        # If the logfile is not writable, inform the user and cancel.
        try:
            open(logfile, 'w').close()
            os.remove(logfile)
        except Exception:
            self.main_window.notify(
                _(
                    "The logfile '%s' is not writable. Please choose "
                    "another location for the logfile."
                ) % logfile
            )
            return None
        if remember_logfile:
            # Remember the logfile folder for the next run
            cfg.default_logfile_folder = os.path.dirname(logfile)
        return logfile

    def get_subject_nr(self, quick=False):
        """Gets the subject number for the current session, either by falling 
        back to a default value of 999 (in quickmode) or through a pop-up 
        dialogue.

        Parameters
        ----------
        quick : bool, optional
            A boolean to indicate whether default should be used for the
            log-file and subject number. Mostly useful while testing the
            experiment.

        Returns
        -------
        int or None
            A subject number or None if no subject number was chosen (i.e. the
            dialogue was cancelled).
        """
        if quick:
            return 999
        subject_nr, ok = QtWidgets.QInputDialog.getInt(
            self.main_window.ui.centralwidget,
            _('Subject number'),
            _('Please enter the subject number'),
            min=0
        )
        if not ok:
            return None
        return subject_nr

    def init_experiment(self, quick=False, fullscreen=False):
        """Initializes a new experiment, which is a newly generated instance of
        the experiment currently active in the user interface.
        
        Parameters
        ----------
        quick : bool, optional
            A boolean to indicate whether default should be used for the
            log-file and subject number. Mostly useful while testing the
            experiment.
        fullscreen : bool, optional
            A boolean to indicate whether the window should be fullscreen.

        Returns
        -------
        bool
            True if the experiment was successfully initiated, False it was
            not.
        """
        # First tell the experiment to get ready, to apply any pending changes,
        # and then initialize the script. This can trigger errors.
        try:
            script = self.main_window.experiment.to_string()
        except Exception as e:
            md = _(
                '# Error\n\nFailed to generate experiment for the '
                'following reason:\n\n- '
            ) + e.markdown()
            self.console.write(e)
            self.tabwidget.open_markdown(md)
            return False
        # Get and set the subject number
        subject_nr = self.get_subject_nr(quick=quick)
        if subject_nr is None:
            return False
        # Get and set the logfile
        logfile = self.get_logfile(quick=quick, subject_nr=subject_nr)
        if logfile is None:
            return False
        # The experiment can be either the full path to the experiment file,
        # the folder of the experiment file, or None.
        if self.main_window.experiment.experiment_path is not None:
            experiment_path = self.main_window.experiment.experiment_path
            if self.main_window.current_path is not None:
                experiment_path = os.path.join(
                    experiment_path,
                    self.main_window.current_path
                )
        else:
            experiment_path = None
        # Build a new experiment. This can trigger a script error.
        try:
            self.experiment = Experiment(
                string=script,
                pool_folder=self.main_window.experiment.pool.folder(),
                experiment_path=experiment_path,
                fullscreen=fullscreen,
                subject_nr=subject_nr,
                logfile=logfile
            )
        except Exception as e:
            md = _('# Error\n\nFailed to parse experiment for the '
                   'following reason:\n\n- ') + safe_str(e)
            self.console.write(e)
            traceback.print_exc()
            self.tabwidget.open_markdown(md)
            return False
        return True

    def run(self, fullscreen=False, quick=False):
        """Runs the experiment.

        Parameters
        ----------
        fullscreen : bool, optional
            A boolean to indicate whether the window should be fullscreen.
        quick : bool, optional
            A boolean to indicate whether default should be used for the
            log-file and subject number. Mostly useful while testing the
            experiment.
        """
        self.main_window.set_run_status('running')
        self.main_window.extension_manager.fire(
            'run_experiment',
            fullscreen=fullscreen
        )
        if not self.init_experiment(quick=quick, fullscreen=fullscreen):
            self.main_window.extension_manager.fire('run_experiment_canceled')
            return
        ret_val = self.execute()
        # PsychoPy deletes the _ built-in
        if '_' not in __builtins__:
            oslogger.warning('re-installing missing gettext built-in')
            import gettext
            gettext.NullTranslations().install()
        self.main_window.set_run_status('finished')
        self.main_window.extension_manager.fire(
            'set_workspace_globals',
            global_dict=self.workspace_globals()
        )
        self.main_window.extension_manager.fire(
            'end_experiment',
            ret_val=ret_val
        )
        if ret_val is None:
            self.main_window.extension_manager.fire(
                'process_data_files',
                data_files=self.data_files()
            )

    def kill(self):
        """Kills an experiment, if the runner supports it."""
        pass

    def workspace_globals(self):
        r"""Returns the experiment's globals dictionary as it was when the
        experiment finished.

        Returns
        -------
        dict
            A globals dictionary.
        """
        return {}

    def data_files(self):

        return self.workspace_globals().get('data_files', [])

    def pause(self):
        r"""Is called when the experiment is paused."""
        self.console.set_workspace_globals(self.workspace_globals())
        print(
            'The experiment has been paused. Switch back to the experiment '
            'window and press space to resume.'
        )
        self.console.show_prompt()
        self.main_window.set_run_status('paused')
        self.main_window.extension_manager.fire('pause_experiment')
        self.paused = True

    def resume(self):
        r"""Is called when the experiment is resumed/ unpaused."""
        self.paused = False
        self.main_window.set_run_status('running')
        self.main_window.extension_manager.fire('resume_experiment')

    @staticmethod
    def has_heartbeat():
        r"""Gives True if the runner supports heartbeats, which are used to
        update the variable inspector, and False otherwise.
        """
        return False



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from libqtopensesame.widgets.header_widget import HeaderWidget
from libqtopensesame.misc.translate import translation_context
_ = translation_context('general_header_widget', category='core')


class GeneralHeaderWidget(HeaderWidget):
    """The widget containing the clickable title and description of the
    experiment.
    
    Parameters
    ----------
    general_tab
        A general_properties object.
    main_window
        A qtopensesame object.
    """
    def __init__(self, general_tab, main_window):
        super().__init__(main_window.experiment)
        self.general_tab = general_tab

    def refresh(self):
        """Updates the header so that it's content match the item."""
        self.set_name(self.experiment.var.title)
        self.set_desc(self.experiment.var.description)
        
    def set_name(self, name, label_type=None):
        """Sets the name.

        Parameters
        ----------
        name
            A name
        """
        super().set_name(name, _('experiment properties'))

    def apply_name(self):
        """Applies the name change and revert the edit control back to the
        static label.
        """
        if self.label_name.isVisible():
            return
        self.label_name.show()
        self.label_type.show()
        self.edit_name.hide()
        self.general_tab.apply_changes()
        self.refresh()

    def apply_desc(self):
        """Applies the description change and revert the edit back to the
        label.
        """
        if self.label_desc.isVisible():
            return
        self.label_desc.show()
        self.edit_desc.hide()
        self.general_tab.apply_changes()
        self.refresh()


# Alias for backwards compatibility
general_header_widget = GeneralHeaderWidget



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from qtpy import QtWidgets


class PluginAction(QtWidgets.QAction):

    """Menu action for a plugin"""
    def __init__(self, main_window, menu, plugin):
        """
        Constructor

        Arguments:
        main_window -- the main window
        menu -- the menu into which the action should be inserted
        plugin -- the name of the plugin
        """
        self.main_window = main_window
        icon = QtGui.QIcon(
            self.main_window.experiment.plugin_manager[plugin].icon_large)
        self.plugin = plugin
        QtWidgets.QAction.__init__(self, icon, "Add %s" % plugin, menu)
        self.triggered.connect(self.add_plugin)

    def add_plugin(self, dummy=None):
        """
        Start a drag to add the plugin to the experiment

        Keyword arguments:
        dummy -- a dummy argument passed by the signal handler (default=None)
        """
        self.main_window.drag_item(self.plugin)


# Alias for backwards compatibility
plugin_action = PluginAction



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from qtpy import QtCore, QtWidgets
from libqtopensesame.widgets.base_widget import BaseWidget
from libqtopensesame.misc.item_name_delegate import ItemNameEditor
from libqtopensesame.misc.translate import translation_context
_ = translation_context('header_widget', category='core')


MAX_NAME_LENGTH = 30


class HeaderWidget(BaseWidget):
    """Editable labels for an item's name and description.
    
    Parameters
    ----------
    item
        QtItem
    """
    def __init__(self, item):
        super().__init__(item.main_window)
        self.setCursor(QtCore.Qt.IBeamCursor)
        self.setToolTip(_("Click to edit"))
        self.item = item
        self.label_name = QtWidgets.QLabel()
        self.label_name.setObjectName('item_name')
        self.label_name.setSizePolicy(QtWidgets.QSizePolicy.Fixed,
                                      QtWidgets.QSizePolicy.Fixed)
        self.label_type = QtWidgets.QLabel()
        self.label_type.setObjectName('item_type')
        self.edit_name = ItemNameEditor()
        self.edit_name.editingFinished.connect(self.apply_name)
        self.edit_name.hide()
        self.label_desc = QtWidgets.QLabel()
        self.label_desc.setObjectName('item_desc')
        self.label_desc.setWordWrap(True)
        self.edit_desc = QtWidgets.QLineEdit()
        self.edit_desc.editingFinished.connect(self.apply_desc)
        self.edit_desc.hide()

        hbox = QtWidgets.QHBoxLayout()
        hbox.setContentsMargins(0, 0, 0, 0)
        hbox.setSpacing(0)
        hbox.addWidget(self.label_name)
        hbox.addWidget(self.label_type)
        name_type = QtWidgets.QWidget()
        name_type.setLayout(hbox)
        vbox = QtWidgets.QVBoxLayout()
        vbox.setContentsMargins(0, 0, 0, 0)
        vbox.setSpacing(0)
        vbox.addWidget(name_type)
        vbox.addWidget(self.edit_name)
        vbox.addWidget(self.label_desc)
        vbox.addWidget(self.edit_desc)
        self.refresh()
        self.setSizePolicy(QtWidgets.QSizePolicy.Expanding,
                           QtWidgets.QSizePolicy.Fixed)
        self.setLayout(vbox)

    def refresh(self):
        """Updates the header so that it's content match the item."""
        self.set_name(self.item.name)
        self.set_desc(self.item.var.description)

    def set_name(self, name, label_type=None):
        """Sets the name.

        Parameters
        ----------
        name: str
        label_type: str or None, optional
        """
        self.label_name.setText(
            name if len(name) < MAX_NAME_LENGTH
            else name[:MAX_NAME_LENGTH - 1] + '…')
        self.label_type.setText(
            ' — ' + label_type if label_type is not None
            else ' — ' + self.item.item_type.replace("_", " "))
        self.edit_name.setText(name)

    def set_desc(self, desc):
        """Sets the description.

        Parameters
        ----------
        name
            A description.
        type
            unicode
        """
        self.edit_desc.setText(safe_decode(desc))
        self.label_desc.setText(safe_decode(desc))

    def apply_name(self):
        """Applies the name change and revert the edit control back to the
        static label.
        """
        if self.label_name.isVisible():
            return
        self.label_name.show()
        self.label_type.show()
        self.edit_name.hide()
        self.item_store.rename(self.item.name, self.edit_name.text())

    def apply_desc(self):
        """Applies the description change and revert the edit back to the
        label.
        """
        if self.label_desc.isVisible():
            return
        description = self.edit_desc.text()
        description = self.item.syntax.sanitize(description)
        self.item.var.description = description
        self.label_desc.setText(description)
        self.label_desc.show()
        self.edit_desc.hide()
        self.item.apply_edit_changes()

    def mousePressEvent(self, event):
        """
        Change the label into an edit for the name or
        the description, depending on where has been
        clicked

        Arguments:
        event -- the mouseClickEvent
        """
        target = self.childAt(event.pos())
        if target == self.label_name:
            self.apply_desc()
            self.label_name.hide()
            self.label_type.hide()
            self.edit_name.show()
            self.edit_name.selectAll()
            self.edit_name.setFocus()
        elif target == self.label_desc:
            self.apply_name()
            self.label_desc.hide()
            self.edit_desc.show()
            self.edit_desc.selectAll()
            self.edit_desc.setFocus()


# Alias for backwards compatibility
header_widget = HeaderWidget



"""A random delay sampled from either a normal or uniform distribution"""

category = "Flow control"
icon = "clock"
controls = [{
    "type": "spinbox",
    "var": "duration",
    "label": "Duration",
    "min_val": 0,
    "max_val": 1000000,
    "name": "spinbox_duration",
    "suffix": " ms",
    "tooltip": "The average duration in milliseconds"
},
    {
    "type": "spinbox",
    "var": "jitter",
    "label": "Jitter",
    "min_val": 0,
    "max_val": 1000000,
    "name": "spinbox_jitter",
    "suffix": " ms",
    "tooltip": "The jitter of the actual duration in milliseconds (depends on Jitter mode)"
},
    {
    "type": "combobox",
    "var": "jitter_mode",
    "label": "Jitter mode",
    "options": [
        "Std. Dev.",
        "Uniform"
    ],
    "name": "combobox_jitter_mode",
    "tooltip": "The mode for determining the actual duration (see Help)"
}
]



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from libopensesame.item_store import ItemStore
from libopensesame import plugins
from libqtopensesame.misc.translate import translation_context
from libopensesame.oslogging import oslogger
_ = translation_context(u'qtitem_store', category=u'core')


class QtItemStore(ItemStore):
    r"""The GUI counterpart of the item store, which also distributes item
    changes etc.
    """
    def __init__(self, experiment):
        r"""Constructor.

        Parameters
        ----------
        main_window : qtopensesame
            The main window object.
        """
        super().__init__(experiment)
        self.error_log = []

    @property
    def main_window(self):
        return self.experiment.main_window

    @property
    def itemtree(self):
        return self.experiment.main_window.itemtree

    @property
    def extension_manager(self):
        return self.main_window.extension_manager
    
    @property
    def plugin_manager(self):
        return self.experiment._plugin_manager

    @property
    def tabwidget(self):
        return self.main_window.tabwidget

    @property
    def console(self):
        return self.main_window.console

    def __delitem__(self, name):
        r"""Deletes an item, and notifies other items of the deletion.

        Parameters
        ----------
        name : str, unicode
            The name of the item to be deleted.
        """
        self.extension_manager.fire(u'prepare_delete_item', name=name)
        # Temporarily disable build_item_tree to avoid recursively calling it,
        # which can take a long time when experiments are big. This is an ugly
        # hack that doesn't fix the recursion issue as such.
        fnc = self.experiment.build_item_tree
        self.experiment.build_item_tree = lambda: None
        del self.__items__[name]
        for _name in self:
            self[_name].remove_child_item(name, index=-1)
        self.experiment.build_item_tree = fnc
        self.experiment.build_item_tree()
        self.extension_manager.fire(u'delete_item', name=name)

    def new(self, _type, name=None, script=None, catch_exceptions=False,
            allow_rename=True):
        """See item_store."""
        import warnings

        with warnings.catch_warnings(record=True) as warning_list:
            if catch_exceptions:
                try:
                    item = super(qtitem_store, self).new(
                        _type=_type,
                        name=name,
                        script=script,
                        allow_rename=allow_rename
                    )
                except Exception as e:
                    self.error_log.append(e)
                    return
            else:
                item = super(qtitem_store, self).new(
                    _type=_type,
                    name=name,
                    script=script,
                    allow_rename=allow_rename
                )
        if warning_list:
            import yaml
            import os
            try:
                self.tabwidget.open_help(
                    os.path.join(u'help', u'new_item_warning'))
            except AttributeError:
                # In case the experiment object doesn't exist yet, which causes
                # the help page to fail.
                pass
            oslogger.warning(yaml.dump(warning_list))
            self.console.write(yaml.dump(warning_list))

        self.main_window.set_unsaved(True)
        return item

    def rename(self, from_name, to_name):
        r"""Renames an item and updates the interface. This function may show a
        notification dialog.

        Parameters
        ----------
        from_name : unicode
            The old item name.
        to_name : unicode
            The desired new item name.

        Returns
        -------
        NoneType, unicode
            The new name of the item, if renaming was successful, None
            otherwise.
        """
        if from_name not in self:
            self.notify(_(u'Item "%s" doesn\'t exist' % from_name))
            return None
        if from_name == to_name:
            return None
        to_name = self.valid_name(
            self[from_name].item_type, suggestion=to_name)
        if to_name in self:
            self.notify(
                _(u'An item with that name already exists.'))
            return None
        if to_name == u'':
            self.notify(_(u'An item name cannot be empty.'))
            return None
        self.extension_manager.fire(u'prepare_rename_item',
                                    from_name=from_name, to_name=to_name)
        # Copy the item in the __items__dictionary
        self.__items__[to_name] = self.__items__[from_name]
        del self.__items__[from_name]
        # Give all items a chance to update
        for item in self.values():
            item.rename(from_name, to_name)
        self.experiment.rename(from_name, to_name)
        self.itemtree.rename(from_name, to_name)
        self.main_window.set_unsaved(True)
        self.extension_manager.fire(u'rename_item', from_name=from_name,
                                    to_name=to_name)
        return to_name

    def set_icon(self, name, icon):
        r"""Changes an item's icon.

        Parameters
        ----------
        name : unicode
            The item name.
        icon : unicode
            The icon name.
        """
        self.itemtree.set_icon(name, icon)

    def used(self):
        """
        returns:
            desc: A list of used item names.
            type: list
        """
        if self.experiment.var.start not in self:
            return []
        return [self.experiment.var.start] + \
                self[self.experiment.var.start].children()

    def unused(self):
        """
        returns:
            desc: A list of unused item names.
            type: list
        """
        _used = self.used()
        return list(filter(lambda item: item not in _used,
                           self.__items__.keys()))

    def valid_type(self, _type):
        """
        arguments:
            _type:
                desc: The item type to check.
                type: str

        returns:
            desc: True if _type is a valid type, False otherwise.
            type: bool
        """
        return _type in self.built_in_types or _type in self.plugin_manager

    def clear_cache(self):
        r"""Clears the cache, currently only the cache with children for each
        item.
        """
        for item in self.values():
            item._children = None

    def is_supported(self, _type):
        supported = self.extension_manager.provide(
            'item_supported', experiment=self.experiment, item_type=_type)
        if supported is not None:
            return supported
        if _type in self.plugin_manager:
            return self.plugin_manager[_type].supports(self.experiment)
        return True

# Alias for backwards compatibility
qtitem_store = QtItemStore



"""This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from qtpy.QtWidgets import QStyledItemDelegate, QLineEdit
from qtpy.QtGui import QFocusEvent
from qtpy.QtCore import Qt, QRect


# Indicates how far the editor should be shifted to the right to make sure it
# doesn't overlap the icon
ICON_MARGIN = 22


class ItemNameEditor(QLineEdit):
    """A custom editor for item names that checks whether the item name ends
    with a '_[item_type]' suffix, and if so, ignores this suffix when selecting
    the text. This encourages users to leave this suffix untouched, thus
    resulting in more meaningful item names.
    """
    def focusInEvent(self, event):
        super().focusInEvent(event)
        item_name = self.text()
        items = self.parent().experiment.items
        # When used in the context of the general properties widget, there is
        # actually no item name, but rather the name of the experiment
        if item_name not in items:
            return
        item_type = self.parent().experiment.items[item_name].item_type
        item_suffix = f'_{item_type}'
        if item_name.endswith(item_suffix):
            self.setSelection(0, len(item_name) - len(item_suffix))


class ItemNameDelegate(QStyledItemDelegate):
    """A delegate that makes uses the custom ItemNameEditor for editing item
    names. The geometry update is required to make sure that the custom 
    editor appears at the correct location, taking into account both the header
    (if present) and the icon.
    """
    def createEditor(self, parent, option, index):
        editor = ItemNameEditor(self.parent())
        text = index.data()
        editor.setText(text)
        return editor

    def updateEditorGeometry(self, editor, option, index):
        r = QRect(option.rect)
        header_height = editor.parent().header().height()
        r.moveTop(r.top() + header_height)
        r.moveLeft(r.left() + ICON_MARGIN)
        editor.setGeometry(r)



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
import os
import time
from qtpy import QtCore
from libopensesame import misc
from libopensesame.oslogging import oslogger
from libqtopensesame.extensions import BaseExtension
from libqtopensesame.misc.config import cfg
from libqtopensesame.misc.translate import translation_context
_ = translation_context(u'automatic_backup', category=u'extension')


class AutomaticBackup(BaseExtension):

    def activate(self):
        if os.name == u"nt":
            os.startfile(self.autosave_folder)
        elif os.name == u"posix":
            misc.open_url(self.autosave_folder)

    def event_startup(self):
        # Create the autosave folder if it doesn't exist yet
        if not os.path.exists(os.path.join(self.main_window.home_folder,
                                           u".opensesame", u"backup")):
            os.mkdir(os.path.join(self.main_window.home_folder, u".opensesame",
                                  u"backup"))
        self.autosave_folder = os.path.join(self.main_window.home_folder,
                                            u".opensesame", u"backup")
        # Remove expired backups
        for path in os.listdir(self.autosave_folder):
            _path = os.path.join(self.autosave_folder, path)
            t = os.path.getctime(_path)
            age = (time.time() - t)/(60*60*24)
            if age > cfg.autosave_max_age:
                oslogger.debug(u"removing '%s'" % path)
                try:
                    os.remove(_path)
                except Exception as e:
                    oslogger.error(f'failed to remove {path}: {e}')
        self.start_autosave_timer()

    def event_run_experiment(self, fullscreen):
        r"""Suspend autosave timer when the experiment starts."""
        if self.autosave_timer is not None:
            oslogger.debug(u"stopping autosave timer")
            self.autosave_timer.stop()

    def event_end_experiment(self, ret_val):
        r"""Resume autosave timer when the experiment ends."""
        if self.autosave_timer is not None:
            oslogger.debug(u"resuming autosave timer")
            self.autosave_timer.start()

    def start_autosave_timer(self):
        r"""Starts the autosave timer."""
        if cfg.autosave_interval > 0:
            oslogger.debug(u"autosave interval = %d ms" %
                           cfg.autosave_interval)
            self.autosave_timer = QtCore.QTimer()
            self.autosave_timer.setInterval(cfg.autosave_interval)
            self.autosave_timer.setSingleShot(True)
            self.autosave_timer.timeout.connect(self.autosave)
            self.autosave_timer.start()
        else:
            oslogger.debug(u"autosave disabled")
            self.autosave_timer = None

    def autosave(self):
        r"""Autosave the experiment if there are unsaved changes."""
        if self.main_window.unsaved_changes:
            path = os.path.join(self.autosave_folder,
                                u'%s.osexp' % str(time.ctime()).replace(u':',
                                                                        u'_'))
            try:
                self.main_window.get_ready()
                self.experiment.save(path, overwrite=True, update_path=False)
                oslogger.debug(u"saving backup as %s" % path)
            except Exception as e:
                oslogger.warning(f'failed to save backup: {e}')
        self.start_autosave_timer()



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from pathlib import Path


class Resources:
    """A central object where resources (package data) can be stored."""
    
    def __init__(self):
        self._folders = [Path(__file__).parent / 'resources']
        self._resources = {}
    
    def add_resource_folder(self, path):
        self._folders.append(Path(path))
    
    def __getitem__(self, key):
        key = key
        if key in self._resources:
            return self._resources[key]
        for folder in self._folders:
            if (folder / key).exists():
                return str(folder / key)
        raise FileNotFoundError(f'failed to find resource {key}')
        
    def __setitem__(self, key, path):
        self._resources[key] = path


resources = Resources()



# coding=utf-8

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from openexp._canvas._gabor.gabor import Gabor
from openexp._canvas._element.psycho import RotatingElement, PsychoElement
from psychopy import visual


class Psycho(RotatingElement, PsychoElement, Gabor):

    def prepare(self):

        env, size = self._mask(self.env, self.size, self.stdev)
        self._stim = visual.GratingStim(
            win=self.win,
            ori=self.orient,
            sf=self.freq,
            phase=self.phase,
            pos=self.to_xy(self.x, self.y),
            tex='sin',
            mask=env,
            size=size,
            color=self.col1.backend_color
        )



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from libopensesame.synth import Synth as SynthRuntime
from libqtopensesame.items.qtplugin import QtPlugin
from libqtopensesame.validators import DurationValidator
from libqtopensesame.misc.translate import translation_context
_ = translation_context('synth', category='item')


class Synth(SynthRuntime, QtPlugin):
    """GUI controls for the synth item."""
    
    description = _('A basic sound synthesizer')
    help_url = 'manual/stimuli/sound'
    lazy_init = True

    def __init__(self, name, experiment, string=None):
        SynthRuntime.__init__(self, name, experiment, string)
        QtPlugin.__init__(self)

    def init_edit_widget(self):
        super().init_edit_widget()
        self.add_combobox_control(
            'osc', _('Waveform'),
            ['sine', 'saw', 'square', 'white_noise'])
        self.add_line_edit_control(
            'freq', _('Frequency'), info=_('In Hertz or as note, e.g. "A1"'))
        self.add_spinbox_control(
            'attack', _('Attack'), min_val=0, max_val=10000000,
            suffix=_(' ms'))
        self.add_spinbox_control(
            'decay', _('Decay'), min_val=0, max_val=10000000, suffix=_(' ms'))
        self.add_doublespinbox_control(
            'volume', _('Volume'), min_val=0, max_val=1,
            suffix=_(' x maximum'))
        self.add_line_edit_control(
            'pan', _('Panning'),
            info=_('Positive values toward the right; "left" or "right" for full panning'))
        self.add_spinbox_control(
            'length', _('Length'), min_val=0, max_val=10000000,
            suffix=_(' ms'))
        self.add_line_edit_control(
            'duration', _('Duration'),
            info=_('In milliseconds, "sound", "keypress", or "mouseclick"'),
            validator=DurationValidator(self, default='sound'))


# Alias for backwards compatibility
synth = Synth



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from libopensesame.oslogging import oslogger
from libopensesame.exceptions import OSException
from libqtopensesame.runners import BaseRunner


class InprocessRunner(BaseRunner):

    """Runs an experiment in the traditional way, in the same process."""
    def execute(self):
        """See base_runner.execute()."""
        # Exceptions during the run phase are important and returned so that the
        # user is notified.
        retval = None
        try:
            self.experiment.run()
        except Exception as e:
            if not isinstance(e, OSException):
                retval = OSException('Unexpected error')
            else:
                retval = e
        # Exceptions during the end phase are less important and only printed
        # to the debug window.
        try:
            self.experiment.end()
        except Exception as _e:
            oslogger.error(u'exception during experiment.end(): %s' % _e)
        return retval

    def workspace_globals(self):
        """See base_runner."""
        if not hasattr(self, u'experiment'):
            return {}
        return self.experiment.python_workspace._globals



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
import os
from libopensesame.exceptions import InvalidOpenSesameScript, OSException, \
    ItemDoesNotExist, LoopSourceFileDoesNotExist, UnsupportedLoopSourceFile, \
    InvalidConditionalExpression, PythonError, InvalidValue, \
    ConditionalExpressionError
from libopensesame.item import Item
from datamatrix import operations, DataMatrix, functional
from pseudorandom import Enforce, MaxRep, MinDist
from openexp.keyboard import Keyboard


class Loop(Item):
    """A loop item runs a single other item multiple times"""
    
    description = u'Repeatedly runs another item'
    valid_orders = u'sequential', u'random'
    commands = [
        u'fullfactorial',
        u'shuffle',
        u'shuffle_horiz',
        u'slice',
        u'sort',
        u'sortby',
        u'reverse',
        u'roll',
        u'weight',
    ]

    def reset(self):
        """See item."""
        self.ef = None
        self.dm = DataMatrix(length=0)
        self.dm.sorted = False
        self.live_dm = None
        self.live_row = None
        self._operations = []
        self._constraints = []
        self._item = u''

        self.var.repeat = 1
        self.var.continuous = u'no'
        self.var.order = u'random'
        self.var.break_if = u'never'
        self.var.break_if_on_first = u'yes'
        self.var.source = u'table'  # file or table
        self.var.source_file = u''

    def from_string(self, string):
        """See item."""
        self.var.clear()
        self.comments = []
        self.reset()
        if string is None:
            return
        for i in string.split(u'\n'):
            self.parse_variable(i)
            cmd, arglist, kwdict = self.syntax.parse_cmd(i)
            if cmd == u'run':
                if len(arglist) != 1 or kwdict:
                    raise InvalidOpenSesameScript('Invalid run command', 
                                                  line=i)
                self._item = arglist[0]
                continue
            if cmd == u'setcycle':
                if self.ef is not None or self._operations:
                    raise InvalidOpenSesameScript(
                        'setcycle must come before constraints and operations',
                        line=i)
                if len(arglist) != 3 or kwdict:
                    raise InvalidOpenSesameScript(
                        'Invalid setcycle command', line=i)
                row, var, val = tuple(arglist)
                if row >= len(self.dm):
                    self.dm.length = row + 1
                if var not in self.dm:
                    self.dm[var] = u''
                self.dm[row][var] = val
                continue
            if cmd == u'constrain':
                if self._operations:
                    raise InvalidOpenSesameScript(
                        'constraints must come before operations', line=i)
                if len(arglist) != 1:
                    raise InvalidOpenSesameScript(
                        'Invalid constrain command', line=i)
                colname = arglist[0]
                for constraint, value in kwdict.items():
                    if constraint == u'maxrep':
                        constraint_cls = MaxRep
                        kwargs = {u'maxrep': value}
                    elif constraint == u'mindist':
                        constraint_cls = MinDist
                        kwargs = {u'mindist': value}
                    else:
                        raise InvalidOpenSesameScript(
                            'Unknown constraint', line=i)
                    self._constraints.append((
                        constraint_cls,
                        colname,
                        kwargs
                    ))
                continue
            if cmd in self.commands:
                self._operations.append((cmd, arglist))
        if len(self.dm) == 0:
            self.dm.length = 1
        if len(self.dm.columns) == 0:
            self.dm.empty_column = u''
        # Backwards compatibility: Older version of OpenSesame can specify the
        # number of cycles through the cycles variable. If the specified number
        # of cycles doesn't match the length of the datamatrix, we change the
        # length of the datamatrix.
        if u'cycles' in self.var and isinstance(self.var.cycles, int) \
                and self.var.cycles != len(self.dm):
            self.dm.length = self.var.cycles

    def to_string(self):
        """See item."""
        # Older versions of OpenSesame use an explicit cycles variable, so we
        # set this here for backwards compatibility.
        self.var.cycles = len(self.dm)
        s = super().to_string()
        for i, row in enumerate(self.dm):
            for name, val in row:
                s += u'\t%s\n' % \
                    self.syntax.create_cmd(u'setcycle', [i, name, val])
        for constraint_cls, colname, kwargs in self._constraints:
            s += u'\t%s\n' % self.syntax.create_cmd(
                u'constrain', [colname], kwargs
            )
        for cmd, arglist in self._operations:
            s += u'\t%s\n' % self.syntax.create_cmd(cmd, arglist)
        s += u'\t%s\n' % self.syntax.create_cmd(u'run', [self._item])
        return s

    def _require_arglist(self, cmd, arglist, minlen=1):
        r"""Checks whether a non-empty argument list has been specified, and
        raises an Exception otherwise.

        Parameters
        ----------
        cmd : str
            The command to check for.
        arglist : list
            The argument list to check.
        """
        if len(arglist) < minlen:
            raise InvalidOpenSesameScript(
                f'Invalid argument list for {cmd}')

    def _create_live_datamatrix(self):
        r"""Builds a live DataMatrix. That is, it takes the original DataMatrix
        and applies all the operations as specified.

        Returns
        -------
        DataMatrix
            A live DataMatrix.
        """
        src_dm = self.dm if self.var.source == u'table' else self._read_file()
        for column_name in src_dm.column_names:
            if not self.syntax.valid_var_name(column_name):
                raise InvalidOpenSesameScript(
                    f'The loop table contains an invalid column name: '
                    f'"{column_name}"')
        # The number of repeats should be numeric. If not, then give an error.
        # This can also occur when generating a preview of a loop table if
        # repeat is variable.
        if not isinstance(self.var.repeat, (int, float)):
            raise InvalidValue(
                f'repeat should be numeric, not {self.var.repeat}')
        length = int(len(src_dm) * self.var.repeat)
        dm = DataMatrix(length=0)
        while len(dm) < length:
            i = min(length-len(dm), len(src_dm))
            if self.var.order == u'random':
                dm <<= operations.shuffle(src_dm)[:i]
            else:
                dm <<= src_dm[:i]
        if self.var.order == u'random':
            dm = operations.shuffle(dm)
        # Constraints come before loop operations
        if self._constraints:
            self.ef = Enforce(dm)
            for constraint_cls, colname, kwargs in self._constraints:
                colname = self.syntax.auto_type(self.syntax.eval_text(colname))
                kwargs = {
                    key: self.syntax.auto_type(self.syntax.eval_text(val))
                    for key, val in kwargs.items()
                }
                try:
                    cols = dm[colname]
                except AttributeError:
                    raise InvalidOpenSesameScript(
                        f'Column {colname} does not exist')
                self.ef.add_constraint(
                    constraint_cls, cols=cols, **kwargs
                )
            dm = self.ef.enforce()
        # Operations come last
        for cmd, arglist in self._operations:
            arglist = [
                self.syntax.auto_type(self.syntax.eval_text(arg))
                for arg in arglist
            ]
            # The column name is always specified last, or not at all
            if arglist:
                try:
                    colname = arglist[-1]
                    if not isinstance(colname, int):
                        col = dm[colname]
                except (IndexError, AttributeError):
                    raise InvalidOpenSesameScript(
                        f'Column {arglist[-1]} does not exist')
            if cmd == u'fullfactorial':
                try:
                    dm = operations.fullfactorial(dm)
                except MemoryError:
                    raise OSException(u'DataMatrix too large for fullfact')
            elif cmd == u'shuffle':
                if not arglist:
                    dm = operations.shuffle(dm)
                else:
                    dm[colname] = operations.shuffle(col)
            elif cmd == u'shuffle_horiz':
                if not arglist:
                    dm = operations.shuffle_horiz(dm)
                else:
                    # There can be multiple column names, so we need to check
                    # if all of them exist, rather than only the last one as
                    # we did above.
                    for _colname in arglist:
                        try:
                            dm[_colname]
                        except AttributeError:
                            raise InvalidOpenSesameScript(
                                f'Column {_colname} does not exist')
                    dm = operations.shuffle_horiz(
                        *[dm[_colname] for _colname in arglist]
                    )
            elif cmd == u'slice':
                self._require_arglist(cmd, arglist, minlen=2)
                dm = dm[arglist[0]: arglist[1]]
            elif cmd == u'sort':
                self._require_arglist(cmd, arglist)
                dm[colname] = operations.sort(col)
            elif cmd == u'sortby':
                self._require_arglist(cmd, arglist)
                dm = operations.sort(dm, by=col)
            elif cmd == u'reverse':
                if not arglist:
                    dm = dm[::-1]
                else:
                    dm[colname] = col[::-1]
            elif cmd == u'roll':
                self._require_arglist(cmd, arglist)
                steps = arglist[0]
                if not isinstance(steps, int):
                    raise InvalidValue('roll steps should be numeric')
                if len(arglist) == 1:
                    dm = dm[-steps:] << dm[:-steps]
                else:
                    dm[colname] = list(col[-steps:]) + list(col[:-steps])
            elif cmd == u'weight':
                self._require_arglist(cmd, arglist)
                # Evaluate the weights before passing them to the weight
                # function, so that weights can be defined in terms of
                # variables.
                ecol = functional.map_(
                    lambda w: self.syntax.auto_type(self.syntax.eval_text(w)),
                    col
                )
                try:
                    dm = operations.weight(ecol)
                except TypeError:
                    raise InvalidValue(
                        'weight values should be non-negative numeric values')
        return dm

    def prepare(self):
        """See item."""
        super().prepare()
        # Compile break-if statement
        break_if = self.var.get(u'break_if', _eval=False)
        if break_if not in (u'never', u''):
            self._break_if = self.syntax.compile_cond(break_if)
        else:
            self._break_if = None
        # Create a keyboard to flush responses between cycles
        self._keyboard = Keyboard(self.experiment)
        # Make sure the item to run exists
        if self._item not in self.experiment.items:
            raise ItemDoesNotExist(self._item)

    def run(self):
        self.set_item_onset()
        if self.live_dm is None or self.var.continuous == u'no':
            self.live_dm = self._create_live_datamatrix()
            self.live_row = 0
        first = True
        while self.live_row < len(self.live_dm):
            self.experiment.var.repeat_cycle = 0
            self.experiment.var.live_row = self.live_row
            self.experiment.var.set('live_row_%s' % self.name, self.live_row)
            for name, val in self.live_dm[self.live_row]:
                # For backwards compatibility, values prefixed by '=' are
                # interpreted as Python code. We rewrite this to f-string
                # notation
                if isinstance(val, str) and val.startswith('='):
                    val = f'{val[1:]}'
                val = self.syntax.auto_type(val,
                                            allow_bool=True,
                                            allow_none=True)
                self.experiment.var.set(name, self.syntax.eval_text(val))
            # Evaluate the run if statement
            if (
                    self._break_if is not None and
                    (not first or self.var.break_if_on_first == u'yes')
            ):
                self.python_workspace[u'self'] = self
                try:
                    if self.python_workspace._eval(self._break_if):
                        break
                except Exception as e:
                    raise ConditionalExpressionError(
                        'Error evaluating break-if expression')
            # Run the item!
            self.experiment.items.execute(self._item)
            # If the repeat_cycle flag was set, run the item again later
            if self.experiment.var.repeat_cycle:
                self.live_dm <<= self.live_dm[self.live_row:self.live_row+1]
                if self.var.order == u'random':
                    self.live_dm = self.live_dm[:self.live_row+1] \
                        << operations.shuffle(self.live_dm[self.live_row+1:])
            self.live_row += 1
            first = False
        else:
            # If the loop finished without breaking, it needs to be reset on
            # the next run of the loop item
            self.live_row = None
            self.live_dm = None

    def _read_file(self):
        r"""Reads a source file and raises an exception if this fails.

        Returns
        -------
        DataMatrix
        """
        from datamatrix import io
        src = self.experiment.pool[self.var.source_file]
        if not os.path.exists(src):
            raise LoopSourceFileDoesNotExist(self.var.source_file)
        if src.endswith(u'.xlsx'):
            try:
                return io.readxlsx(src)
            except Exception as e:
                raise UnsupportedLoopSourceFile(
                    f'Failed to read .xlsx file: {src}')
        try:
            return io.readtxt(src)
        except Exception as e:
            raise UnsupportedLoopSourceFile(
                f'Failed to read text file: {src}. Perhaps it has the wrong '
                f'format or it is not utf-8 encoded')

    def _var_info_table(self):
        """
        returns:
                A list of (var, value) tuples that have been defined in the loop
                table.
        """
        return [(colname, safe_decode(col)) for colname, col in self.dm.columns]

    def _var_info_file(self):
        """
        returns:
                A list of (var, value) tuples that have been defined in a source
                file.
        """
        try:
            dm = self._read_file()
        except Exception:
            return []
        return [(colname, safe_decode(col)) for colname, col in dm.columns]

    def var_info(self):
        """See item."""
        return super().var_info() + (
            self._var_info_table()
            if self.var.source == u'table'
            else self._var_info_file()
        )


# Alias for backwards compatibility
loop = Loop



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from libqtopensesame.items.qtplugin import QtPlugin
from libqtopensesame import validators
from libqtopensesame.misc.translate import translation_context
_ = translation_context('qtautoplugin', category='core')


class QtAutoPlugin(QtPlugin):

    """A class that processes auto-plugins defined in a YAML file"""
    def init_edit_widget(self):
        """Construct the GUI controls based on info.yaml"""
        super().init_edit_widget(False)
        item_type_translate = translation_context(self.item_type,
                                                  category='plugin')
        # The help url should be specified as `help_url`. However, old-style
        # plugins used `help`.
        self.help_url = self.plugin_attribute('help_url')
        if self.help_url is None:
            self.help_url = self.plugin_attribute('help')
        # Some options are required. Which options are requires depends on the
        # specific widget.
        required = [
            (['checkbox', 'color_edit', 'combobox', 'editor', 'filepool',
              'line_edit', 'spinbox', 'text'], ['label']),
            (['checkbox', 'color_edit', 'combobox', 'editor', 'filepool',
              'line_edit', 'spinbox'], ['var']),
            (['spinbox', 'slider'], ['min_val', 'max_val']),
            (['combobox'], ['options']),
        ]
        # Keywords are optional parameters that are set to some default if they
        # are not specified.
        keywords = {
            'info': None,
            'min_width': None,
            'prefix': '',
            'suffix': '',
            'left_label': 'min.',
            'right_label': 'max.',
            'syntax': False,
            'language': 'python'
        }
        # This indicates whether we should pad the controls with a stretch at
        # the end.
        need_stretch = True
        for c in self.plugin_attribute('controls'):
            # Check whether all required options have been specified
            if 'type' not in c:
                raise RuntimeError(
                    _('You must specify "type" for %s controls in info.yaml')
                    % option)
            for types, options in required:
                if c['type'] in types:
                    for option in options:
                        if option not in c:
                            raise RuntimeError(
                                _('You must specify "%s" for %s controls in info.yaml')
                                % (option, c['type']))
            if 'var' in c and not self.syntax.valid_var_name(c['var']):
                raise RuntimeError(
                    _('Invalid variable name (%s) specified in %s plugin info') %
                    (c['var'], self.item_type))
            # Set missing keywords to None
            for keyword, default in keywords.items():
                if keyword not in c:
                    c[keyword] = default
            # Translate translatable fields
            c['label'] = item_type_translate(c['label'])
            if c['info'] is not None:
                c['info'] = item_type_translate(c['info'])
            # Parse checkbox
            if c['type'] == 'checkbox':
                widget = self.add_checkbox_control(c['var'], c['label'],
                                                   info=c['info'])
            # Parse color_edit
            elif c['type'] == 'color_edit':
                widget = self.add_color_edit_control(c['var'], c['label'],
                    info=c['info'], min_width=c['min_width'])
            # Parse combobox
            elif c['type'] == 'combobox':
                widget = self.add_combobox_control(c['var'], c['label'],
                    c['options'], info=c['info'])
            # Parse editor
            elif c['type'] == 'editor':
                widget = self.add_editor_control(c['var'], c['label'],
                    syntax=c['syntax'], language=c['language'])
                need_stretch = False
            # Parse filepool
            elif c['type'] == 'filepool':
                widget = self.add_filepool_control(c['var'], c['label'],
                    info=c['info'])
            # Parse line_edit
            elif c['type'] == 'line_edit':
                widget = self.add_line_edit_control(c['var'], c['label'],
                    info=c['info'], min_width=c['min_width'])
            # Parse line_edit
            elif c['type'] == 'conditional_expression':
                widget = self.add_conditional_expression_control(
                    c['var'], c['label'], c['verb'], info=c['info'],
                    min_width=c['min_width'])
            # Parse spinbox
            elif c['type'] == 'spinbox':
                widget = self.add_spinbox_control(c['var'], c['label'],
                    c['min_val'], c['max_val'], prefix=c['prefix'],
                    suffix=c['suffix'], info=c['info'])
            # Parse slider
            elif c['type'] == 'slider':
                widget = self.add_slider_control(c['var'], c['label'],
                    c['min_val'], c['max_val'], left_label=c['left_label'],
                    right_label=c['right_label'], info=c['info'])
            # Parse text
            elif c['type'] == 'text':
                widget = self.add_text(c['label'])
            else:
                raise Exception(_('"%s" is not a valid qtautoplugin control')
                                % controls['type'])
            # Add an optional validator
            if 'validator' in c:
                try:
                    validator = getattr(validators,
                                        '%s_validator' % c['validator'])
                except:
                    raise RuntimeError(
                        'Invalid validator: %s' % c['validator'])
                widget.setValidator(validator(self.main_window))
            # Add the widget as an item property when the 'name' option is
            # specified.
            if 'name' in c:
                if hasattr(self, c['name']):
                    raise RuntimeError(
                        _('Name "%s" is already taken in qtautoplugin control')
                        % c['name'])
                setattr(self, c['name'], widget)
        if need_stretch:
            self.add_stretch()

    def apply_edit_changes(self):
        """Applies the controls. I.e. sets the variables from the controls."""
        if not super().apply_edit_changes() or self.lock:
            return False
        return True

    def edit_widget(self):
        """Sets the controls based on the variables."""
        self.lock = True
        super().edit_widget()
        self.lock = False
        return self._edit_widget


# Alias for backwards compatibility
qtautoplugin = QtAutoPlugin
